# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""ALTER TABLE tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query().create_table(tbl).column('id', 'INTEGER').primary_key().column('num', 'INTEGER'),
    )


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='describe unimplemented')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='describe uninplemented')),
    ],
    indirect=True,
)
async def test_rename_to(txn: Transaction) -> None:
    """Test alter table statement."""
    tbl = Table('tbl')
    dst = Table('ref')

    assert await txn.exq(Query().alter_table(tbl).rename_to(dst)) == []
    assert await txn.exq(Query().pragma('table_info', value='ref')) == [
        (0, 'id', 'INTEGER', 0, None, 1),
        (1, 'num', 'INTEGER', 0, None, 0),
    ]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='describe unimplemented')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='describe uninplemented')),
    ],
    indirect=True,
)
async def test_rename_column(txn: Transaction) -> None:
    """Test alter table statement."""
    tbl = Table('tbl')
    assert await txn.exq(Query().alter_table(tbl).rename_column(tbl.id, tbl.tid)) == []
    assert await txn.exq(Query().pragma('table_info', value='tbl')) == [
        (0, 'tid', 'INTEGER', 0, None, 1),
        (1, 'num', 'INTEGER', 0, None, 0),
    ]


@pytest.mark.skipif('sys.version_info < (3, 9)')
@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='describe unimplemented')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='describe uninplemented')),
    ],
    indirect=True,
)
async def test_drop_column(txn: Transaction) -> None:
    """Test alter table statement."""
    tbl = Table('tbl')
    assert await txn.exq(Query().alter_table(tbl).drop_column(tbl.num)) == []
    assert await txn.exq(Query().pragma('table_info', value='tbl')) == [
        (0, 'id', 'INTEGER', 0, None, 1),
    ]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='describe unimplemented')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='describe uninplemented')),
    ],
    indirect=True,
)
async def test_add_column(txn: Transaction) -> None:
    """Test alter table statement."""
    tbl = Table('tbl')

    assert await txn.exq(Query().alter_table(tbl).add_column(tbl.name, 'TEXT')) == []
    assert await txn.exq(Query().pragma('table_info', value='tbl')) == [
        (0, 'id', 'INTEGER', 0, None, 1),
        (1, 'num', 'INTEGER', 0, None, 0),
        (2, 'name', 'TEXT', 0, None, 0),
    ]

    assert (
        await txn.exq(
            Query()
            .alter_table(tbl)
            .add_column(tbl.x1, 'INTEGER')
            .references(tbl, tbl.id)
            .on_update()
            .cascade()
            .deferrable()
            .default(1)
            .not_null()
            .on_conflict_ignore(),
        )
        == []
    )
    assert await txn.exq(Query().pragma('table_info', value='tbl')) == [
        (0, 'id', 'INTEGER', 0, None, 1),
        (1, 'num', 'INTEGER', 0, None, 0),
        (2, 'name', 'TEXT', 0, None, 0),
        (3, 'x1', 'INTEGER', 1, '1', 0),
    ]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Database connection and transaction management."""

from .exceptions import IntegrityError, OperationalError
from .manager import get_manager

__all__ = [
    'IntegrityError',
    'OperationalError',
    'get_manager',
]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL VACUUM."""

from .base import QBase, Term
from .expr import Expr, ExprLiteral


class Vacuum(QBase[tuple[()]]):
    """VACUUM mixin."""

    def vacuum(self, schema: str | None = None, into: Expr | str | None = None) -> Term[tuple[()]]:
        """Create VACUUM statement.

        Args:
            schema: Schema to run on or all.
            into: Target database to vacuum into.

        Returns:
            Query.

        """
        if schema:
            schema = schema.replace('"', '""')
            schema = f' "{schema}"'
        else:
            schema = ''

        if into:
            if not isinstance(into, Expr):
                into = ExprLiteral(into)
            sql, args = into.__getsql__()
            sql = f' INTO {sql}'
        else:
            sql = ''
            args = ()

        sql = f'VACUUM{schema}{sql}'
        return self._f(Term, sql, args)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Model field types."""

from __future__ import annotations

from datetime import date, datetime, time
from typing import Annotated

AutoField = Annotated[int, 'AutoField']
BigAutoField = Annotated[int, 'BigAutoField']
IntegerField = Annotated[int, 'IntegerField']
BigIntegerField = Annotated[int, 'BigIntegerField']
SmallIntegerField = Annotated[int, 'SmallIntegerField']
FloatField = Annotated[float, 'FloatField']
DoubleField = Annotated[float, 'DoubleField']
DecimalField = Annotated[float, 'DecimalField']  # max_digits, decimals
CharField = Annotated[str, 'CharField']  # max_length
FixedCharField = Annotated[str, 'FixedCharField']  # length
TextField = Annotated[str, 'TextField']
BlobField = Annotated[bytes, 'BlobField']
UUIDField = Annotated[str, 'UUIDField']
DateTimeField = Annotated[datetime, 'DateTimeField']
DateField = Annotated[date, 'DateField']
TimeField = Annotated[time, 'TimeField']
BooleanField = Annotated[bool, 'BooleanField']
ForeignKeyField = Annotated[int, 'ForeignKeyField']  # model, on_delete



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Model query generators."""

from __future__ import annotations

from dataclasses import fields
from itertools import combinations
from types import NoneType
from typing import TYPE_CHECKING, Annotated, Union, get_args, get_origin, get_type_hints

from sqls.queries import Query, Table

from .model import Fieldmeta, Model

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TypeVar

    from sqls.interfaces import Query as QueryProtocol

    T = TypeVar('T')

SQLITE_TYPEMAP = {
    'AutoField': 'INTEGER',
    'BigAutoField': 'INTEGER',
    'IntegerField': 'INTEGER',
    'BigIntegerField': 'INTEGER',
    'SmallIntegerField': 'INTEGER',
    'FloatField': 'REAL',
    'DoubleField': 'REAL',
    'DecimalField': 'DECIMAL',
    'CharField': 'VARCHAR',
    'FixedCharField': 'CHAR',
    'TextField': 'TEXT',
    'BlobField': 'BLOB',
    'UUIDField': 'TEXT',
    'DateTimeField': 'DATETIME',
    'DateField': 'DATE',
    'TimeField': 'TIME',
    'BooleanField': 'INTEGER',
    'ForeignKeyField': 'INTEGER',
}


def get_create_queries(  # noqa: C901, PLR0912, PLR0915
    models: Sequence[type[Model]],
) -> list[QueryProtocol[tuple[()]]]:
    """Get create table queries for models."""
    query: QueryProtocol[tuple[()]]
    queries: list[QueryProtocol[tuple[()]]] = []
    idxqueries: list[QueryProtocol[tuple[()]]] = []

    byname = {model.__model__['name']: model for model in models}

    for name, model in byname.items():
        tablemeta = model.__tablemeta__
        modelmeta = model.__model__
        table = Table(tablemeta.name or name, tablemeta.schema)

        query = Query().create_table(table, if_not_exists=True)

        hints = get_type_hints(model, include_extras=True)
        haspk = False
        for sub in fields(model):
            nullable = False
            fieldmeta = Fieldmeta()
            hint = hints[sub.name]
            if hint is NoneType:
                continue

            if get_origin(hint) is Union:
                assert get_args(hint)[1] is NoneType
                nullable = True
                hint = get_args(hint)[0]

            assert get_origin(hint) == Annotated
            hint, *hintargs = get_args(hint)
            argtype = hintargs[0]
            if len(hintargs) == 2:
                fieldmeta = hintargs[1]

            suffix = f'({fieldmeta.max_length})' if fieldmeta.max_length is not None else ''
            query = query.column(sub.name, f'{SQLITE_TYPEMAP[argtype]}{suffix}')
            if sub.name == 'id':
                query = query.primary_key(autoincrement=True).not_null()
                haspk = True
            if argtype == 'ForeignKeyField':
                if fieldmeta.foreign_key:
                    foreign_key = fieldmeta.foreign_key
                else:
                    assert sub.name.endswith('_id')
                    foreign_key = (sub.name[:-3], 'id')
                modelmeta['temp_rels'].append((haspk, *foreign_key, sub.name))
                other = Table(foreign_key[0])
                if sub.name != 'id' and not nullable:
                    query = query.not_null()
                query = query.references(other, getattr(other, foreign_key[1]))
                if fieldmeta.on_delete == 'RESTRICT':
                    query = query.on_delete().restrict()
                else:
                    query = query.on_delete().cascade()
                if sub.name != 'id':
                    idxtable = Table(f'idx__{name}__{sub.name}')
                    idxqueries.append(
                        Query().create_index(
                            idxtable,
                            table,
                            getattr(table, sub.name),
                            if_not_exists=True,
                        ),
                    )
            else:
                if not nullable:
                    query = query.not_null()
                if fieldmeta.unique:
                    query = query.unique()
                if fieldmeta.default is not None:
                    query = query.default(fieldmeta.default)

        queries.append(query)

        for indexmeta in tablemeta.indexes:
            idxtable = Table(f'idx__{name}__{"__".join(indexmeta.fields)}')
            idxqueries.append(
                Query().create_index(
                    idxtable,
                    table,
                    *[getattr(table, x) for x in indexmeta.fields],
                    unique=indexmeta.unique,
                    if_not_exists=True,
                ),
            )

    for model in models:
        model.__model__['forward_rels'].clear()
        model.__model__['backward_rels'].clear()
        model.__model__['through_rels'].clear()

    for model in models:
        modelmeta = model.__model__
        indirects = []
        for direct, othername, backfield, fwdfield in modelmeta['temp_rels']:
            if direct:
                othermodel = byname[othername]

                model.__model__['forward_rels'].append((othermodel, backfield, fwdfield))
                othermodel.__model__['backward_rels'].append((model, fwdfield, backfield))
            else:
                indirects.append((othername, backfield, fwdfield))

        for left, right in combinations(indirects, 2):
            leftmodel = byname[left[0]]
            rightmodel = byname[right[0]]
            leftmodel.__model__['through_rels'].append(
                (model, left[2], left[1], rightmodel, *right[1:]),
            )
            rightmodel.__model__['through_rels'].append(
                (model, right[2], right[1], leftmodel, *left[1:]),
            )

        modelmeta['temp_rels'].clear()

    return queries + idxqueries



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL SELECT."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, cast, overload

from .base import QBase, T_co, Term
from .exceptions import QueryError
from .expr import Expr, ExprAlias, ExprLiteral, ExprSubq, ExprTyped
from .table import TableQuery, Tabular

if TYPE_CHECKING:
    from sqls.interfaces import BindArg, SqlValue

    T_forward = TypeVar('T_forward', bound='tuple[SqlValue, ...]')
    T_query = TypeVar('T_query', bound='QBase[tuple[SqlValue, ...]]')

    _T1 = TypeVar('_T1', bound='SqlValue | object')
    _T2 = TypeVar('_T2', bound='SqlValue | object')
    _T3 = TypeVar('_T3', bound='SqlValue | object')
    _T4 = TypeVar('_T4', bound='SqlValue | object')
    _T5 = TypeVar('_T5', bound='SqlValue | object')
    _T6 = TypeVar('_T6', bound='SqlValue | object')
    _T7 = TypeVar('_T7', bound='SqlValue | object')
    _T8 = TypeVar('_T8', bound='SqlValue | object')
    _T9 = TypeVar('_T9', bound='SqlValue | object')
    _T10 = TypeVar('_T10', bound='SqlValue | object')
    _T11 = TypeVar('_T11', bound='SqlValue | object')
    _T12 = TypeVar('_T12', bound='SqlValue | object')
    _T13 = TypeVar('_T13', bound='SqlValue | object')
    _T14 = TypeVar('_T14', bound='SqlValue | object')
    _T15 = TypeVar('_T15', bound='SqlValue | object')
    _T16 = TypeVar('_T16', bound='SqlValue | object')


class Select(QBase[tuple[()]]):
    """SELECT mixin."""

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        __col12: ExprTyped[_T12],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        __col12: ExprTyped[_T12],
        __col13: ExprTyped[_T13],
        /,
        distinct: bool = False,
    ) -> PostSelect[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13]]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        __col12: ExprTyped[_T12],
        __col13: ExprTyped[_T13],
        __col14: ExprTyped[_T14],
        /,
        distinct: bool = False,
    ) -> PostSelect[
        tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13, _T14],
    ]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        __col12: ExprTyped[_T12],
        __col13: ExprTyped[_T13],
        __col14: ExprTyped[_T14],
        __col15: ExprTyped[_T15],
        /,
        distinct: bool = False,
    ) -> PostSelect[
        tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13, _T14, _T15],
    ]: ...

    @overload
    def select(
        self,
        __col1: ExprTyped[_T1],
        __col2: ExprTyped[_T2],
        __col3: ExprTyped[_T3],
        __col4: ExprTyped[_T4],
        __col5: ExprTyped[_T5],
        __col6: ExprTyped[_T6],
        __col7: ExprTyped[_T7],
        __col8: ExprTyped[_T8],
        __col9: ExprTyped[_T9],
        __col10: ExprTyped[_T10],
        __col11: ExprTyped[_T11],
        __col12: ExprTyped[_T12],
        __col13: ExprTyped[_T13],
        __col14: ExprTyped[_T14],
        __col15: ExprTyped[_T15],
        __col16: ExprTyped[_T16],
        /,
        distinct: bool = False,
    ) -> PostSelect[
        tuple[
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
        ],
    ]: ...

    def select(  # type: ignore[misc]
        self,
        *columns: ExprTyped[SqlValue | object],
        distinct: bool = False,
    ) -> PostSelect[tuple[SqlValue, ...]]:
        """Start SELECT statement.

        Args:
            columns: Result columns.
            distinct: Is distinct query.

        Returns:
            Query builder.

        """
        pad = '' if not self._parent else ' '
        distinctstr = ' DISTINCT' if distinct else ''
        cols = []
        args: list[BindArg] = []
        for col in columns:
            sql, argz = col.__getsql__()
            cols.append(sql)
            args += argz
        sql = f'{pad}SELECT{distinctstr} {", ".join(cols)}'
        return self._f(PostSelect, sql, tuple(args))

    def values(self, *vals: T_forward) -> PostSelect[T_forward]:
        """Start VALUES clause.

        Args:
            vals: Result expressions.

        Returns:
            Query builder.

        Raises:
            QueryError: If values empty or not nested sequence.

        """
        if not vals:
            msg = 'Values requires at least one row'
            raise QueryError(msg)

        if not vals[0]:
            msg = 'Values row requires at least one column'
            raise QueryError(msg)

        pad = '' if not self._parent else ' '
        row = f'({", ".join("?" * len(vals[0]))})'
        placeholder = ','.join([row] * len(vals))
        return self._f(PostSelect, f'{pad}VALUES{placeholder}', tuple(y for x in vals for y in x))


class From(QBase[T_co]):
    """FROM mixin."""

    def from_(self, table: Tabular) -> PostFrom[T_co]:
        """Add FROM clause.

        Args:
            table: Input table.

        Returns:
            Query builder.

        """
        assert isinstance(table, Expr | ExprAlias | Tabular), f'From param is not tabular, {table}'
        sql, args = table.__getsql__()
        return self._f(PostFrom, f' FROM {sql}', args)


class Where(QBase[T_co]):
    """WHERE mixin."""

    def where(self, expr: Expr | None) -> PostWhere[T_co]:
        """Add WHERE clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        if expr is None:
            return PostWhere(self)
        sql, args = expr.__getsql__()
        return self._f(PostWhere, f' WHERE {sql}', args)


class Groupby(QBase[T_co]):
    """GROUP BY mixin."""

    def group_by(self, *exprs: Expr) -> PostGroupby[T_co]:
        """Add GROUP BY clause.

        Args:
            exprs: Expressions.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for expr in exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        return self._f(PostGroupby, f' GROUP BY {", ".join(sqls)}', tuple(args))


class Having(QBase[T_co]):
    """HAVING mixin."""

    def having(self, expr: Expr) -> PostHaving[T_co]:
        """Add HAVING clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostHaving, f' HAVING {sql}', args)


class Compound(QBase[T_co]):
    """Compund mixin."""

    def union(self) -> PostCompound:
        """Add UNION query."""
        return self._f(PostCompound, ' UNION')

    def union_query(self, other: T_query) -> T_query:
        """Add UNION query.

        Args:
            other: Select statement.

        Returns:
            Query builder.

        """
        assert isinstance(other, QBase)
        copy = self._f(other.__class__, ' UNION ')
        copy._append(other)  # noqa: SLF001
        return cast('T_query', copy)

    def union_all(self) -> PostCompound:
        """Add UNION ALL query."""
        return self._f(PostCompound, ' UNION ALL')

    def intersect(self) -> PostCompound:
        """Add INTERSECT query."""
        return self._f(PostCompound, ' INTERSECT')

    def except_(self) -> PostCompound:
        """Add EXCEPT query."""
        return self._f(PostCompound, ' EXCEPT')


class Order(QBase[T_co]):
    """ORDER mixin."""

    def order_by(
        self,
        *exprs: Expr,
        desc: bool = False,
        nulls: str | None = None,
    ) -> PostOrder[T_co]:
        """Add ORDER BY clause.

        Args:
            exprs: Expressions.
            desc: Sort descending.
            nulls: NULLS keyword to add.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for expr in exprs:
            sql, argz = expr.__getsql__()
            sqls.append(sql)
            args += argz
        sort = ' DESC' if desc else ''
        nulls = '' if nulls is None else f' NULLS {nulls}'
        return self._f(PostOrder, f' ORDER BY {", ".join(sqls)}{sort}{nulls}', tuple(args))


class Limit(QBase[T_co]):
    """LIMIT mixin."""

    def limit(self, expr: int | Expr) -> PostLimit[T_co]:
        """Add LIMIT clause.

        Args:
            expr: Item limit.

        Returns:
            Query builder.

        """
        if not isinstance(expr, Expr):
            expr = ExprLiteral(expr)
        sql, args = expr.__getsql__()
        return self._f(PostLimit, f' LIMIT {sql}', args)


class Offset(QBase[T_co]):
    """OFFSET mixin."""

    def offset(self, expr: int | Expr) -> PostOffset[T_co]:
        """Add OFFSET clause.

        Args:
            expr: Item offset.

        Returns:
            Query builder.

        """
        if not isinstance(expr, Expr):
            expr = ExprLiteral(expr)
        sql, args = expr.__getsql__()
        return self._f(PostOffset, f' OFFSET {sql}', args)


class SelectTerm(Term[T_co]):
    """Select terminal state mixin."""

    def subq(self) -> ExprSubq[T_co]:
        """Convert to subquery expression."""
        return ExprSubq(self)

    def table(self) -> TableQuery:
        """Convert to subquery table."""
        return TableQuery(self)


class PostSelect(
    From[T_co],
    Where[T_co],
    Groupby[T_co],
    Having[T_co],
    Compound[T_co],
    Order[T_co],
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post select builder."""


class PostFrom(
    Where[T_co],
    Groupby[T_co],
    Having[T_co],
    Compound[T_co],
    Order[T_co],
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post from builder."""


class PostWhere(
    Groupby[T_co],
    Having[T_co],
    Compound[T_co],
    Order[T_co],
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post where builder."""


class PostGroupby(
    Having[T_co],
    Compound[T_co],
    Order[T_co],
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post groupby builder."""


class PostHaving(
    Compound[T_co],
    Order[T_co],
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post having builder."""


class PostCompound(
    Select,
):
    """Post compound builder."""


class PostOrder(
    Limit[T_co],
    SelectTerm[T_co],
):
    """Post order builder."""


class PostLimit(
    Offset[T_co],
    SelectTerm[T_co],
):
    """Post limit builder."""


class PostOffset(
    SelectTerm[T_co],
):
    """Post offset builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Sqlite manager tests."""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, cast
from unittest.mock import AsyncMock, patch

import pytest

from sqls.transactions import IntegrityError, OperationalError, get_manager

from .common import Query

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from sqls.interfaces import Manager
    from sqls.transactions.sqlite import Manager as SqliteManager


@pytest.fixture()
async def manager() -> AsyncGenerator[Manager, None]:
    """In memory database manager."""
    res = get_manager('file:memdb1?mode=memory&cache=shared')
    await res.init()
    yield res
    await res.close()


async def test_options() -> None:
    """Test options are set on manager."""
    res = cast(
        'SqliteManager',
        get_manager(
            'file:memdb1?mode=memory&cache=shared',
            {
                'max_connections': 2,
                'init_pragmas': ['PRAGMA foo'],
                'fini_pragmas': ['PRAGMA bar'],
            },
        ),
    )
    assert res.max_connections == 2
    with patch('aiosqlite.Connection.execute', new_callable=AsyncMock) as execute:
        await res.init()
        async with res.txn():
            pass
        assert execute.call_count == 4
        assert execute.call_args.args == ('PRAGMA foo',)
        await res.close()
        assert execute.call_count == 5
        assert execute.call_args.args == ('PRAGMA bar',)


async def test_manager(manager: SqliteManager) -> None:
    """Test Manager."""
    assert len(manager.pool) == 0

    async with manager.txn() as txn:
        assert txn
    assert len(manager.pool) == 1

    async with manager.txn() as txn, manager.txn() as inner_txn:
        assert inner_txn != txn
    assert len(manager.pool) == 2

    async with manager.txn() as txn:
        await txn.execute('SELECT 1')
        assert await txn.exq_count(Query('SELECT 1')) == 1
        assert await txn.exq(Query('SELECT 1')) == [(1,)]

    with pytest.raises(IntegrityError, match='UNIQUE'):  # noqa: PT012
        async with manager.txn(True) as txn:
            await txn.exq(Query('CREATE TABLE foo (bar INTEGER PRIMARY KEY)'))
            assert await txn.exq_count(Query('INSERT INTO foo (bar) VALUES(1),(2)')) == 2
            await txn.exq(Query('INSERT INTO foo (bar) VALUES(1)'))

    with pytest.raises(OperationalError, match='syntax error'):
        async with manager.txn(True) as txn:
            await txn.exq(Query('BAD SYNTAX'))

    async with manager.txn(True) as txn:
        await txn.exq(Query('CREATE TABLE foo (bar INTEGER PRIMARY KEY)'))
        with suppress(ValueError):
            async with txn as innertxn:
                await innertxn.exq(Query('INSERT INTO foo (bar) VALUES(1)'))
                raise ValueError
        async with txn as innertxn:
            assert await innertxn.exq(Query('INSERT INTO foo (bar) VALUES(1)')) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL CREATE TRIGGER."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from sqls.interfaces import BindArg

    from .expr import Expr, ExprColumn
    from .table import Table


class CreateTrigger(QBase[tuple[()]]):
    """CREATE TRIGGER mixin."""

    def create_trigger(
        self,
        trigger: Table,
        temp: bool = False,
        if_not_exists: bool = False,
    ) -> PostCreate:
        """Start CREATE TRIGGER statement.

        Args:
            trigger: Table.
            temp: Is trigger temporary.
            if_not_exists: Ignore if trigger exists.

        Returns:
            Query builder.

        """
        sql, _ = trigger.__getsql__()
        tmp = ' TEMP' if temp else ''
        ine = ' IF NOT EXISTS' if if_not_exists else ''
        sql = f'CREATE{tmp} TRIGGER{ine} {sql}'
        return self._f(PostCreate, sql)


class Moment(QBase[tuple[()]]):
    """Moment mixin."""

    def before(self) -> PostMoment:
        """Add BEFORE keyword."""
        return self._f(PostMoment, ' BEFORE')

    def after(self) -> PostMoment:
        """Add AFTER keyword."""
        return self._f(PostMoment, ' AFTER')

    def instead_of(self) -> PostMoment:
        """Add INSTEAD OF keyword."""
        return self._f(PostMoment, ' INSTEAD OF')


class Cause(QBase[tuple[()]]):
    """Cause mixin."""

    def delete(self) -> PostCause:
        """Add DELETE keyword."""
        return self._f(PostCause, ' DELETE')

    def insert(self) -> PostCause:
        """Add INSERT keyword."""
        return self._f(PostCause, ' INSERT')

    def update(self, of_: list[ExprColumn] | None = None) -> PostCause:
        """Add UPDATE OF keyword.

        Args:
            of_: List of columns.

        Returns:
            Query builder.

        """
        sql = f' OF {", ".join(x.col for x in of_)}' if of_ else ''
        return self._f(PostCause, f' UPDATE{sql}')


class On(QBase[tuple[()]]):
    """ON mixin."""

    def on_(self, table: Table) -> PostOn:
        """Add ON clause.

        Args:
            table: Table to trigger on.

        Returns:
            Query builder.

        """
        sql, args = table.__getsql__()
        return self._f(PostOn, f' ON {sql}', args)


class Foreach(QBase[tuple[()]]):
    """FOR EACH ROW mixin."""

    def for_each_row(self) -> PostForeach:
        """FOR EACH ROW clause."""
        return self._f(PostForeach, ' FOR EACH ROW')


class When(QBase[tuple[()]]):
    """WHEN mixin."""

    def when(self, expr: Expr) -> PostWhen:
        """Add WHEN clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostWhen, f' WHEN {sql}', args)


class Statements(QBase[tuple[()]]):
    """Trigger statements mixin."""

    def statements(self, *statements: QBase[tuple[()]]) -> PostStatements:
        """Add trigger statements.

        Args:
            statements: Queries to run in trigger.

        Returns:
            Query.

        """
        sqls = []
        args: list[BindArg] = []
        for statement in statements:
            sql, argz = statement.__getsql__()
            sqls.append(f'{sql};')
            args += argz
        return self._f(PostStatements, f' BEGIN {" ".join(sqls)} END', tuple(args))


class PostCreate(
    Moment,
    Cause,
):
    """Post create builder."""


class PostMoment(
    Cause,
):
    """Post moment builder."""


class PostCause(
    On,
):
    """Post cause builder."""


class PostOn(
    Foreach,
    When,
    Statements,
):
    """Post on builder."""


class PostForeach(
    When,
    Statements,
):
    """Post for each builder."""


class PostWhen(
    Statements,
):
    """Post when builder."""


class PostStatements(
    Term[tuple[()]],
):
    """Post statements builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""PRAGMA tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_pragma(txn: Transaction) -> None:
    """Test pragma statement."""
    tbl = Table('foo')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))

    query = Query().pragma('encoding')
    assert query.__getsql__() == ('PRAGMA "encoding"', ())
    assert await txn.exq(query) == [('UTF-8',)]

    query = Query().pragma('table_info', schema='main', value='foo')
    assert query.__getsql__() == ('PRAGMA "main"."table_info" = \'foo\'', ())
    assert await txn.exq(query) == [(0, 'id', 'INTEGER', 0, None, 0)]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL Query Builder."""

from .exceptions import QueryError
from .explain import Explain
from .expr import ExprFunction, ExprLiteral
from .query import Query
from .table import Join, Table
from .window import Window

__all__ = [
    'Explain',
    'ExprFunction',
    'ExprLiteral',
    'Join',
    'Query',
    'QueryError',
    'Table',
    'Window',
]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL EXPLAIN."""

from .query import Query


class Explain(Query):
    """EXPLAIN Query Builer."""

    def __init__(
        self,
        query_plan: bool | None = False,
    ) -> None:
        """Initialize.

        Args:
            query_plan: Get query plan.

        """
        super().__init__()
        sql = ' QUERY PLAN' if query_plan else ''
        self._parts += [f'EXPLAIN{sql} ']



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL PRAGMA."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from sqls.interfaces import SqlValue


class Pragma(QBase[tuple[()]]):
    """PRAGMA mixin."""

    def pragma(
        self,
        pragma: str,
        schema: str | None = None,
        value: str | float | None = None,  # | int
    ) -> Term[tuple[SqlValue, ...]]:
        """Create PRAGMA statement.

        Args:
            pragma: Pragma name or function with value.
            schema: Schema name.
            value: Value to set.

        Returns:
            Query.

        """
        if schema:
            schema = schema.replace('"', '""')
            schema = f'"{schema}".'
        else:
            schema = ''

        value = f' = {value!r}' if value else ''

        pragma = pragma.replace('"', '""')
        return self._f(Term, f'PRAGMA {schema}"{pragma}"{value}')



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""REINDEX tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table
from sqls.transactions.sqlite import Transaction as SqliteTransaction

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_simple(txn: Transaction) -> None:
    """Test reindex statement."""
    query = Query().reindex()
    assert query.__getsql__() == ('REINDEX', ())
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_table(txn: Transaction) -> None:
    """Test reindex statement."""
    tbl = Table('tbl')
    assert await txn.exq(Query().create_table(tbl).column('id', 'INTEGER')) == []

    query = Query().reindex('tbl')
    assert query.__getsql__() == ('REINDEX "tbl"', ())
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_schema_table(txn: Transaction) -> None:
    """Test reindex statement."""
    schema = 'main' if isinstance(txn, SqliteTransaction) else 'public'
    tbl = Table('tbl', schema=schema)
    assert await txn.exq(Query().create_table(tbl).column('id', 'INTEGER')) == []

    query = Query().reindex(tbl)
    assert query.__getsql__() == (f'REINDEX "{schema}"."tbl"', ())
    assert await txn.exq(query) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL CREATE INDEX."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from .expr import Expr, ExprColumn
    from .table import Table


class CreateIndex(QBase[tuple[()]]):
    """CREATE INDEX mixin."""

    def create_index(
        self,
        index: Table,
        table: Table,
        *cols: ExprColumn,
        unique: bool = False,
        if_not_exists: bool = False,
    ) -> PostCreate:
        """Start CREATE INDEX statement.

        Args:
            index: New index table.
            table: Table to index over.
            cols: Columns to index.
            unique: Should index be unique.
            if_not_exists: Ignore if index exists.

        Returns:
            Query.

        """
        isql, _ = index.__getsql__()
        tsql, _ = table.__getsql__()

        def colname(col: ExprColumn) -> str:
            return f'"{col.col}"'

        colstr = f'({", ".join(colname(x) for x in cols)})'
        unq = ' UNIQUE' if unique else ''
        ine = ' IF NOT EXISTS' if if_not_exists else ''
        sql = f'CREATE{unq} INDEX{ine} {isql} ON {tsql} {colstr}'
        return self._f(PostCreate, sql)


class Where(QBase[tuple[()]]):
    """WHERE mixin."""

    def where(self, expr: Expr) -> PostWhere:
        """Add WHERE clause.

        Args:
            expr: Expression.

        Returns:
            Query.

        """
        sql, args = expr.__getsql__()
        return self._f(PostWhere, f' WHERE {sql}', args)


class PostCreate(
    Where,
    Term[tuple[()]],
):
    """Post CREATE mixin."""


class PostWhere(
    Term[tuple[()]],
):
    """Post WHERE mixin."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Queries test fixtures."""

from __future__ import annotations

import warnings
from os import getenv
from typing import TYPE_CHECKING

import pytest

from sqls.transactions import get_manager

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from sqls.interfaces import Transaction


@pytest.fixture(params=['sqlite', 'postgres', 'mysql'])
async def txn(request: pytest.FixtureRequest) -> AsyncGenerator[Transaction, None]:
    """Transaction fixture."""
    if request.param == 'sqlite':
        manager = get_manager(':memory:')
        async with manager.txn(True) as transaction:
            yield transaction
        await manager.close()
    elif request.param == 'postgres':
        host = getenv('SQLS_POSTGRES_HOST', 'localhost')
        manager = get_manager(f'postgresql://sqlsuser:sqlspass@{host}/sqls')
        await manager.init()
        async with manager.pool.acquire() as connection:  # type: ignore[attr-defined]
            await connection.fetch('DROP SCHEMA public CASCADE')
            await connection.fetch('CREATE SCHEMA public')
        async with manager.txn() as transaction:
            yield transaction
        await manager.close()
    elif request.param == 'mysql':
        host = getenv('SQLS_MYSQL_HOST', 'localhost')
        manager = get_manager(f'mysql://sqlsuser:sqlspass@{host}:3306/sqls')
        await manager.init()
        warnings.filterwarnings('ignore', module='aiomysql')
        async with manager.txn() as transaction:
            await transaction.execute('DROP TABLE IF EXISTS `tbl` CASCADE')
            await transaction.execute('DROP TABLE IF EXISTS `dst` CASCADE')
            await transaction.execute('DROP VIEW IF EXISTS `vw_` CASCADE')
        async with manager.txn() as transaction:
            yield transaction
        await manager.close()
        warnings.filterwarnings('default', module='aiomysql')



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""DROP tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query().create_table(tbl).column('id', 'INTEGER').primary_key(),
    )


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='syntax not implemented')),
    ],
    indirect=True,
)
async def test_drop_index(txn: Transaction) -> None:
    """Test drop statements."""
    tbl = Table('tbl')
    idx = Table('idx')
    await txn.exq(Query().create_index(idx, tbl, tbl.id))

    query = Query().drop_index(idx)
    assert query.__getsql__()[0] == 'DROP INDEX "idx"'
    await txn.exq(query)

    query = Query().drop_index(idx, if_exists=True)
    assert query.__getsql__()[0] == 'DROP INDEX IF EXISTS "idx"'
    await txn.exq(query)


async def test_drop_table(txn: Transaction) -> None:
    """Test drop statements."""
    tbl = Table('tbl')
    query = Query().drop_table(tbl)
    assert query.__getsql__()[0] == 'DROP TABLE "tbl"'
    await txn.exq(query)

    query = Query().drop_table(tbl, if_exists=True)
    assert query.__getsql__()[0] == 'DROP TABLE IF EXISTS "tbl"'
    await txn.exq(query)


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_drop_trigger(txn: Transaction) -> None:
    """Test drop statements."""
    tbl = Table('tbl')
    trig = Table('trig')
    assert (
        await txn.exq(
            Query()
            .create_trigger(trig)
            .delete()
            .on_(tbl)
            .statements(
                Query().insert(tbl).columns(tbl.id).values((tbl.id,)),
            ),
        )
        == []
    )
    query = Query().drop_trigger(trig)
    assert query.__getsql__()[0] == 'DROP TRIGGER "trig"'
    await txn.exq(query)

    query = Query().drop_trigger(trig, if_exists=True)
    assert query.__getsql__()[0] == 'DROP TRIGGER IF EXISTS "trig"'
    await txn.exq(query)


async def test_drop_view(txn: Transaction) -> None:
    """Test drop statements."""
    tbl = Table('tbl')
    view = Table('vw_')
    await txn.exq(Query().create_view(view).as_(Query().select(tbl.id.typed(int)).from_(tbl)))

    query = Query().drop_view(view)
    assert query.__getsql__()[0] == 'DROP VIEW "vw_"'
    await txn.exq(query)

    query = Query().drop_view(view, if_exists=True)
    assert query.__getsql__()[0] == 'DROP VIEW IF EXISTS "vw_"'
    await txn.exq(query)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Squal tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_analyze(txn: Transaction) -> None:
    """Test analyze statement."""
    tbl = Table('tbl')
    await txn.exq(Query().create_table(tbl).column('id', 'INTEGER'))

    query = Query().analyze()
    assert query.__getsql__() == ('ANALYZE', ())
    assert await txn.exq(query) == []

    query = Query().analyze(tbl)
    assert query.__getsql__() == ('ANALYZE "tbl"', ())
    assert await txn.exq(query) == []

    query = Query().analyze('tbl')
    assert query.__getsql__() == ('ANALYZE "tbl"', ())
    assert await txn.exq(query) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL CREATE DROP."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term

if TYPE_CHECKING:
    from .table import Table


class Drop(QBase[tuple[()]]):
    """DROP mixin."""

    def drop_index(self, index: Table, if_exists: bool = False) -> PostDrop:
        """Create DROP INDEX statement.

        Args:
            index: Index to drop.
            if_exists: Execute statement only if index exists.

        Returns:
            Query.

        """
        sql, _ = index.__getsql__()
        ife = ' IF EXISTS' if if_exists else ''
        return self._f(PostDrop, f'DROP INDEX{ife} {sql}')

    def drop_table(self, table: Table, if_exists: bool = False) -> PostDrop:
        """Create DROP TABLE statement.

        Args:
            table: Table to drop.
            if_exists: Execute statement only if table exists.

        Returns:
            Query.

        """
        sql, _ = table.__getsql__()
        ife = ' IF EXISTS' if if_exists else ''
        return self._f(PostDrop, f'DROP TABLE{ife} {sql}')

    def drop_trigger(self, trigger: Table, if_exists: bool = False) -> PostDrop:
        """Create DROP TRIGGER statement.

        Args:
            trigger: Trigger to drop.
            if_exists: Execute statement only if trigger exists.

        Returns:
            Query.

        """
        sql, _ = trigger.__getsql__()
        ife = ' IF EXISTS' if if_exists else ''
        return self._f(PostDrop, f'DROP TRIGGER{ife} {sql}')

    def drop_view(self, view: Table, if_exists: bool = False) -> PostDrop:
        """Create DROP VIEW statement.

        Args:
            view: Viewt o drop.
            if_exists: Execute statement only if view exists.

        Returns:
            Query.

        """
        sql, _ = view.__getsql__()
        ife = ' IF EXISTS' if if_exists else ''
        return self._f(PostDrop, f'DROP VIEW{ife} {sql}')


class PostDrop(
    Term[tuple[()]],
):
    """Post drop builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Transactions test tools."""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from sqls.interfaces import BindArg

T = TypeVar('T')


class Query(Generic[T]):
    """Query helper."""

    attr: T

    def __init__(self, query: str, *args: BindArg) -> None:
        """Initialize."""
        self.query = query
        self.args = args

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Retrieve query string."""
        return self.query, self.args



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL INSERT."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import QBase, Term
from .exceptions import QueryError
from .expr import Expr, ExprLiteral

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sqls.interfaces import BindArg, SqlValue

    from .base import Texplicit
    from .expr import ExprColumn
    from .table import Tabular


class Insert(QBase[tuple[()]]):
    """INSERT mixin."""

    def _insert(self, cmd: str, table: Tabular) -> PostInsert:
        """Start 'action' INTO helper.

        Args:
            cmd: Action.
            table: Target table.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()
        return self._f(PostInsert, f'{cmd} INTO {sql}')

    def insert(self, table: Tabular) -> PostInsert:
        """Start INSERT INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT', table)

    def replace(self, table: Tabular) -> PostInsert:
        """Start REPLACE INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('REPLACE', table)

    def insert_or_replace(self, table: Tabular) -> PostInsert:
        """Start INSERT OR REPLACE INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT OR REPLACE', table)

    def insert_or_rollback(self, table: Tabular) -> PostInsert:
        """Start INSERT OR ROLLBACK INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT OR ROLLBACK', table)

    def insert_or_abort(self, table: Tabular) -> PostInsert:
        """Start INSERT OR ABORT statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT OR ABORT', table)

    def insert_or_fail(self, table: Tabular) -> PostInsert:
        """Start INSERT OR FAIL INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT OR FAIL', table)

    def insert_or_ignore(self, table: Tabular) -> PostInsert:
        """Start INSERT OR IGNORE INTO statement.

        Args:
            table: Target table.

        Returns:
            Query builder.

        """
        return self._insert('INSERT OR IGNORE', table)


class Columns(QBase[tuple[()]]):
    """Columns mixin."""

    def columns(self, *cols: ExprColumn) -> PostColumns:
        """Add target columns.

        Args:
            cols: Columns.

        Returns:
            Query builder.

        """
        return self._f(PostColumns, f""" ({", ".join(f'"{x.col}"' for x in cols)})""")

    def default_values(self) -> PostValues:
        """Add DEFAULT VALUES clause."""
        return self._f(PostValues, ' DEFAULT VALUES')


class Values(QBase[tuple[()]]):
    """Values mixin."""

    def values(self, *vals: tuple[SqlValue | Expr, ...] | ExprLiteral) -> PostValues:
        """Add VALUES clause.

        Args:
            vals: Tuples of column values.

        Returns:
            Query builder.

        Raises:
            QueryError: If values empty or not nested sequence.

        """
        if not vals:
            msg = 'Values requires at least one row'
            raise QueryError(msg)

        if not vals[0]:
            msg = 'Values row requires at least one column'
            raise QueryError(msg)

        sqls = []
        args: list[BindArg] = []
        for val in vals:
            valex = val if isinstance(val, Expr) else ExprLiteral(val)
            sql, argz = valex.__getsql__()
            sqls.append(sql)
            args += argz
        return self._f(PostValues, f' VALUES{",".join(sqls)}', tuple(args))

    def select(self, select: QBase[Texplicit]) -> PostValues:
        """Add select query.

        Args:
            select: Select statement.

        Returns:
            Query builder.

        """
        sql, args = select.__getsql__()
        return self._f(PostValues, f' {sql}', args)


class Upsert(QBase[tuple[()]]):
    """UPSERT mixin."""

    def on_conflict(self, *cols: ExprColumn) -> PostUpsertColumns:
        """Start ON CONFLICT clause.

        Args:
            cols: Columns.

        Returns:
            Query builder.

        """
        sqls = []
        args: list[BindArg] = []
        for col in cols:
            sql, argz = col.__getsql__()
            sqls.append(sql)
            args += argz
        return self._f(PostUpsertColumns, f' ON CONFLICT ({", ".join(sqls)}) DO', tuple(args))

    def on_conflict_do_nothing(self) -> PostUpsert:
        """Add ON CONFLICT DO NOTHING clause."""
        return self._f(PostUpsert, ' ON CONFLICT DO NOTHING')


class UpsertUpdate(QBase[tuple[()]]):
    """UPDATE SET mixin."""

    def update_set(
        self,
        cols: Iterable[ExprColumn],
        vals: Iterable[Expr | int],
    ) -> PostUpsertUpdate:
        """Add UPDATE SET clause.

        Args:
            cols: Columns.
            vals: Values.

        Returns:
            Query builder.

        Raises:
            QueryError: If column and value lengths differ.

        """
        cols = tuple(cols)
        vals = tuple(vals)
        if not cols or len(cols) != len(vals):
            msg = 'Columns and values need to have same length.'
            raise QueryError(msg)

        sqls = []
        args: list[BindArg] = []
        for col, val in zip(cols, vals, strict=True):
            valex = val if isinstance(val, Expr) else ExprLiteral(val)
            sql, argz = valex.__getsql__()
            args += argz
            sqls.append(f'{col.col} = {sql}')
        return self._f(PostUpsertUpdate, f' UPDATE SET {", ".join(sqls)}', tuple(args))


class UpsertWhere(QBase[tuple[()]]):
    """WHERE mixin."""

    def where(self, expr: Expr) -> PostUpsertWhere:
        """Add WHERE clause.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostUpsertWhere, f' WHERE {sql}', args)


class PostInsert(
    Columns,
):
    """Post insert builder."""


class PostColumns(
    Values,
):
    """Post columns builder."""


class PostValues(
    Upsert,
    Term[tuple[()]],
):
    """Post values builder."""


class PostUpsert(
    Term[tuple[()]],
):
    """Post upsert builder."""


class PostUpsertColumns(
    UpsertUpdate,
):
    """Post upsert columns builder."""


class PostUpsertUpdate(
    UpsertWhere,
    Term[tuple[()]],
):
    """Post upsert update builder."""


class PostUpsertWhere(
    Term[tuple[()]],
):
    """Post upsert where builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Expression tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprFunction, ExprLiteral, Query
from sqls.queries.expr import Expr, ExprColumn, ExprTyped

if TYPE_CHECKING:
    from typing import TypeVar

    from sqls.interfaces import Transaction

    T = TypeVar('T')


async def evl(txn: Transaction, expr: ExprTyped[T]) -> T:
    """Evaluate expression."""
    query = Query().select(expr)
    res = await txn.exq(query)
    return res[0][0]


async def test_simple(txn: Transaction) -> None:
    """Test expressions."""
    with pytest.raises(NotImplementedError):
        Expr().__getsql__()

    nonbind = ExprLiteral("fo'o", bind=False)
    assert nonbind.__getsql__() == ("'fo''o'", ())
    assert await evl(txn, nonbind.typed(str)) == "fo'o"

    nonbind = ExprLiteral((1, 'x'), bind=False)
    assert nonbind.__getsql__() == ("(1, 'x')", ())

    assert str(ExprLiteral(1.0)) == 'Literal: 1.0'


@pytest.mark.skip()
async def test_unsupported() -> None:
    """Test expressions."""
    nonbind = ExprLiteral(((1, 2, 3), ExprLiteral((4, 5, 6))))  # type: ignore[arg-type]
    assert nonbind.__getsql__() == ('(?, (?, ?, ?))', ((1, 2, 3), 4, 5, 6))


async def test_column() -> None:
    """Test expressions."""

    def format_col(scheme: str, table: str, col: str) -> str:
        return ExprColumn(scheme, table, col).__getsql__()[0]

    assert format_col('', '', '*') == '*'
    assert format_col('', 'tbl', 'col') == '"tbl"."col"'
    assert format_col('scheme', 'tbl', 'col') == '"scheme"."tbl"."col"'
    assert format_col('', 'tbl', 'co"l') == '"tbl"."co""l"'

    assert isinstance(ExprColumn('', 'tbl', 'col').typed(int), ExprTyped)


async def test_positive_negative(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1)
    oneplusone = one + 1

    assert await evl(txn, one.positive().typed(int)) == 1
    assert await evl(txn, one.negative().typed(int)) == -1
    assert await evl(txn, oneplusone.negative().typed(int)) == -2


async def test_negate(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1)
    true = ExprLiteral(True)

    assert await evl(txn, one.bitwise_negate().typed(int)) in [-2, 18446744073709551614]

    assert await evl(txn, true.negate().typed(int)) in [False, 0]


async def test_strconcat(txn: Transaction) -> None:
    """Test expressions."""
    strone = ExprLiteral('1.0')
    foobar = ExprLiteral('foo_bar')

    assert await evl(txn, strone.strconcat('foo').typed(str)) == '1.0foo'
    assert await evl(txn, strone.strconcat(foobar).typed(str)) == '1.0foo_bar'


async def test_arithmetic(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)
    oneint = ExprLiteral(1)
    oneplusone = oneint + 1

    assert await evl(txn, (one * 2).typed(int)) == 2
    assert await evl(txn, (one * oneplusone).typed(int)) == 2
    assert await evl(txn, (one / 2).typed(float)) == 0.5
    assert await evl(txn, (one / oneplusone).typed(float)) == 0.5
    assert await evl(txn, (one + 2).typed(int)) == 3
    assert await evl(txn, (one + oneplusone).typed(int)) == 3
    assert await evl(txn, (one - 2).typed(int)) == -1
    assert await evl(txn, (one - oneplusone).typed(int)) == -1
    assert await evl(txn, (oneint << 2).typed(int)) == 4
    assert await evl(txn, (oneint << oneplusone).typed(int)) == 4
    assert await evl(txn, (oneint * 15 >> 2).typed(int)) == 3
    assert await evl(txn, (oneint * 15 >> oneplusone).typed(int)) == 3
    assert await evl(txn, (oneint * 15 & 2).typed(int)) == 2
    assert await evl(txn, (oneint * 15 & oneplusone).typed(int)) == 2
    assert await evl(txn, (oneint * 16 | 2).typed(int)) == 18
    assert await evl(txn, (oneint * 16 | oneplusone).typed(int)) == 18

    assert await evl(txn, (one < 2).typed(bool)) == 1
    assert await evl(txn, (one < oneplusone).typed(bool)) == 1
    assert await evl(txn, (one <= 2).typed(bool)) == 1
    assert await evl(txn, (one <= oneplusone).typed(bool)) == 1
    assert await evl(txn, (one > 2).typed(bool)) == 0
    assert await evl(txn, (one > oneplusone).typed(bool)) == 0
    assert await evl(txn, (one >= 2).typed(bool)) == 0
    assert await evl(txn, (one >= oneplusone).typed(bool)) == 0
    assert await evl(txn, (one == 2).typed(bool)) == 0
    assert await evl(txn, (one == oneplusone).typed(bool)) == 0
    assert await evl(txn, (one != 2).typed(bool)) == 1
    assert await evl(txn, (one != oneplusone).typed(bool)) == 1


async def test_modulo(txn: Transaction) -> None:
    """Test expressions."""
    oneint = ExprLiteral(1)
    oneplusone = oneint + 1

    assert await evl(txn, (oneint * 5 % 2).typed(bool)) == 1
    assert await evl(txn, (oneint * 5 % oneplusone).typed(bool)) == 1


async def test_boolean_algebra(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, (one != 2).and_(True).typed(bool)) == 1
    assert await evl(txn, (one != 2).and_(False).typed(bool)) == 0
    assert await evl(txn, (one != 2).and_(one == 2).typed(bool)) == 0
    assert await evl(txn, (one == 2).or_(True).typed(bool)) == 1
    assert await evl(txn, (one == 2).or_(False).typed(bool)) == 0
    assert await evl(txn, (one == 2).or_(one != 2).typed(bool)) == 1


async def test_cast(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, (one == 2).cast('TEXT').typed(str)) in ['false', '0']


async def test_like(txn: Transaction) -> None:
    """Test expressions."""
    foobar = ExprLiteral('foo_bar')

    assert await evl(txn, foobar.like(foobar).typed(bool)) == 1
    assert await evl(txn, foobar.like('foo%').typed(bool)) == 1
    assert await evl(txn, foobar.not_like('foo').typed(bool)) == 1

    assert await evl(txn, foobar.like('fo___ar').typed(bool)) == 1
    assert await evl(txn, foobar.like('foo_bar').typed(bool)) == 1
    assert await evl(txn, foobar.like('_oo!_ba_', escape='!').typed(bool)) == 1


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_glob(txn: Transaction) -> None:
    """Test expressions."""
    foobar = ExprLiteral('foo_bar')

    assert await evl(txn, foobar.glob('foo*').typed(bool)) == 1
    assert await evl(txn, foobar.not_glob('*foo').typed(bool)) == 1


@pytest.mark.parametrize(
    'txn',
    [
        pytest.param('sqlite', marks=pytest.mark.skip(reason='needs c-extension')),
        'postgres',
        'mysql',
    ],
    indirect=True,
)
async def test_regex(txn: Transaction) -> None:
    """Test expressions."""
    foobar = ExprLiteral('foo_bar')

    assert await evl(txn, foobar.regexp('^foo.*$').typed(bool)) == 1
    assert await evl(txn, foobar.not_regexp('^foo$').typed(bool)) == 1


@pytest.mark.parametrize(
    'txn',
    [
        pytest.param('sqlite', marks=pytest.mark.skip(reason='raises by default')),
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_match(txn: Transaction) -> None:
    """Test expressions."""
    foobar = ExprLiteral('foo_bar')

    assert await evl(txn, foobar.match('foo*').typed(bool)) == 1
    assert await evl(txn, foobar.not_match('*foo').typed(bool)) == 1


async def test_null(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)
    null = ExprLiteral(None)

    assert await evl(txn, one.is_null().typed(bool)) == 0
    assert await evl(txn, one.not_null().typed(bool)) == 1

    assert await evl(txn, null.is_null().typed(bool)) == 1
    assert await evl(txn, null.not_null().typed(bool)) == 0

    assert await evl(txn, null.is_(null).typed(bool)) == 1
    assert await evl(txn, null.is_not(null).typed(bool)) == 0


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_null_is(txn: Transaction) -> None:
    """Test expressions."""
    null = ExprLiteral(None)

    assert await evl(txn, null.is_(0).typed(bool)) == 0
    assert await evl(txn, null.is_not(0).typed(bool)) == 1


async def test_between(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, (one + 7).between(4, 12).typed(bool)) == 1
    assert await evl(txn, (one + 7).between(one + 3, one + 11).typed(bool)) == 1
    assert await evl(txn, one.not_between(2, 3).typed(bool)) == 1
    assert await evl(txn, one.not_between(one + 1, one + 2).typed(bool)) == 1


async def test_in(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, one.in_((1, 3, 5)).typed(bool)) == 1
    assert await evl(txn, one.in_(ExprLiteral((1, 3, 5))).typed(bool)) == 1
    assert await evl(txn, one.not_in((2, 4, 6)).typed(bool)) == 1
    assert await evl(txn, one.not_in(ExprLiteral((2, 4, 6))).typed(bool)) == 1


async def test_subq_exists(txn: Transaction) -> None:
    """Test expressions."""
    assert await evl(txn, Query().values((1,)).subq().exists().typed(int)) == 1
    assert await evl(txn, Query().values((1,)).subq().not_exists().typed(int)) == 0


async def test_alias(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, (one + 10).as_('foo').typed(int)) == 11


async def test_function(txn: Transaction) -> None:
    """Test expressions."""
    one = ExprLiteral(1.0)

    assert await evl(txn, ExprFunction('SUM', one).typed(int)) == 1



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Transaction manager."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sqls.interfaces import Manager, ManagerOptions


def get_manager(uri: str, options: ManagerOptions | None = None) -> Manager:
    """Get manager instance for URI."""
    if uri.startswith('mysql:'):
        from .mysql import Manager as MysqlManager

        return MysqlManager(uri, options)
    if uri.startswith('postgresql:'):
        from .postgresql import Manager as PostgresqlManager

        return PostgresqlManager(uri, options)
    assert uri.startswith(('sqlite:', 'file:', ':memory:'))
    from .sqlite import Manager as SqliteManager

    return SqliteManager(uri, options)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQLite manager."""

from __future__ import annotations

import asyncio
import sqlite3
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, cast

import aiosqlite

from .exceptions import IntegrityError, OperationalError

if TYPE_CHECKING:
    import sys
    from collections.abc import AsyncGenerator
    from types import TracebackType
    from typing import TypeVar

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from sqls.interfaces import BindArg, ManagerOptions, Query, SqlValue, SupportsGetsql

    T = TypeVar('T')


class Transaction:
    """Transaction class."""

    def __init__(self, connection: Connection) -> None:
        """Initialzize transaction object."""
        self.connection = connection

    async def execute(
        self,
        query: str,
        *args: BindArg,
    ) -> list[tuple[SqlValue, ...]]:
        """Execute query."""
        cursor = self.connection.cursor
        assert cursor
        await cursor.execute(query, args)
        return await cursor.fetchall()  # type: ignore[return-value]

    async def exq(self, query: Query[T]) -> list[T]:
        """Execute shorthand."""
        cursor = self.connection.cursor
        assert cursor
        try:
            assert (
                self.connection.immediate
                or 'SELECT' in query.__getsql__()[0]
                or 'ATTACH' in query.__getsql__()[0]
            )
            await cursor.execute(*query.__getsql__())
        except sqlite3.IntegrityError as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise IntegrityError(msg) from None
        except sqlite3.OperationalError as err:
            msg = f'QUERY: {query.__getsql__()!r}, ERROR {err.args!r}.'
            raise OperationalError(msg) from None

        return await cursor.fetchall()  # type: ignore[return-value]

    async def exq_count(self, query: SupportsGetsql) -> int:
        """Execute count shorthand."""
        cursor = self.connection.cursor
        assert cursor
        assert self.connection.immediate or 'SELECT' in query.__getsql__()[0]
        await cursor.execute(*query.__getsql__())
        if cursor.rowcount == -1:
            return len(await cursor.fetchall())  # type: ignore[arg-type]
        return cursor.rowcount

    async def __aenter__(self) -> Self:
        """Transaction setup."""
        assert self.connection.cursor
        await self.connection.cursor.execute(f'SAVEPOINT "{id(self)}"')
        return self.__class__(self.connection)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        """Transaction cleanup."""
        if exc is not None:
            await self.connection.connection.execute(
                f'ROLLBACK TRANSACTION TO SAVEPOINT "{id(self)}"',
            )
            raise exc
        await self.connection.connection.execute(f'RELEASE SAVEPOINT "{id(self)}"')
        return False


class Connection:
    """Sqlite connection."""

    def __init__(self, connection: aiosqlite.Connection) -> None:
        """Initialize."""
        self.connection = connection
        self.cursor: aiosqlite.Cursor | None = None
        self.immediate = False

    async def close(self) -> None:
        """Close connection."""
        await self.connection.close()

    async def __aenter__(self) -> Transaction:
        """Transaction setup."""
        self.cursor = await self.connection.cursor()
        await self.cursor.execute(f'BEGIN {"IMMEDIATE " if self.immediate else ""}TRANSACTION')
        return Transaction(self)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        """Transaction cleanup."""
        assert self.cursor
        if exc is not None:
            await self.cursor.execute('ROLLBACK TRANSACTION')
            await self.cursor.close()
            self.cursor = None
            raise exc
        await self.cursor.execute('COMMIT TRANSACTION')
        await self.cursor.close()
        return False


class Manager:
    """Database manager."""

    def __init__(self, uri: str, options: ManagerOptions | None = None) -> None:
        """Initialize."""
        self.uri = uri
        if not options:
            options = cast('ManagerOptions', {})
        self.max_connections = options.get('max_connections', 4)
        self.init_pragmas = options.get('init_pragmas', [])
        self.fini_pragmas = options.get('fini_pragmas', [])

        self.queue: asyncio.Queue[Connection] = asyncio.Queue()
        self.pool: list[Connection | None] = []

    async def init(self) -> None:
        """Initialize pool."""

    async def get_connection(self) -> Connection:
        """Create or retrieve a connection object."""
        if not self.queue.qsize() and len(self.pool) < self.max_connections:
            index = len(self.pool)
            self.pool.append(None)
            connection = Connection(await aiosqlite.connect(self.uri, uri=True))
            await connection.connection.execute('PRAGMA foreign_keys = ON')
            await connection.connection.execute('PRAGMA journal_mode = WAL')
            await connection.connection.execute('PRAGMA synchronous = normal')
            for pragma in self.init_pragmas:
                await connection.connection.execute(pragma)
            self.pool[index] = connection
            return connection

        return await self.queue.get()

    async def close(self) -> None:
        """Close all database connections."""
        for connection in self.pool:
            assert connection
            for pragma in self.fini_pragmas:
                await connection.connection.execute(pragma)
            await connection.close()

    @asynccontextmanager
    async def txn(self, immediate: bool = False) -> AsyncGenerator[Transaction, None]:
        """Open a transaction for database operations."""
        connection = await self.get_connection()
        connection.immediate = immediate
        try:
            async with connection as newtxn:
                yield newtxn
                newtxn.connection = None  # type: ignore[assignment]
        finally:
            await self.queue.put(connection)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL DETACH."""

from .base import QBase, Term


class Detach(QBase[tuple[()]]):
    """DETACH DATABASE mixin."""

    def detach(self, schema: str) -> Term[tuple[()]]:
        """Create DETACH DATABASE statement.

        Args:
            schema: Schema name to deattach.

        Returns:
            Query.

        """
        schema = schema.replace('"', '""')
        sql = f'DETACH DATABASE "{schema}"'
        return self._f(Term, sql)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""INSERT tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, QueryError, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('name', 'TEXT')
        .unique()
        .column('age', 'INTEGER'),
    )


async def test_default_values(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    query = Query().insert(tbl).default_values()
    assert query.__getsql__()[0] == 'INSERT INTO "tbl" DEFAULT VALUES'
    assert await txn.exq(query) == []


async def test_insert_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    with pytest.raises(QueryError, match='one row'):
        Query().insert(tbl).columns(tbl.name, tbl.age).values().__getsql__()

    with pytest.raises(QueryError, match='one column'):
        Query().insert(tbl).columns(tbl.name, tbl.age).values(()).__getsql__()

    query = Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42))
    assert query.__getsql__() == ('INSERT INTO "tbl" ("name", "age") VALUES(?, ?)', ('Ringo', 42))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        'mysql',
    ],
    indirect=True,
)
async def test_replace_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().replace(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == ('REPLACE INTO "tbl" ("name", "age") VALUES(?, ?)', ('Ringo', 43))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_insert_or_replace_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().insert_or_replace(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == (
        'INSERT OR REPLACE INTO "tbl" ("name", "age") VALUES(?, ?)',
        ('Ringo', 43),
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_insert_or_rollback_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().insert_or_rollback(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == (
        'INSERT OR ROLLBACK INTO "tbl" ("name", "age") VALUES(?, ?)',
        ('Ringo', 43),
    )
    with pytest.raises(Exception, match='UNIQUE constraint failed'):
        await txn.exq(query)

    await txn.execute('BEGIN TRANSACTION')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_insert_or_abort_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().insert_or_abort(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == (
        'INSERT OR ABORT INTO "tbl" ("name", "age") VALUES(?, ?)',
        ('Ringo', 43),
    )
    with pytest.raises(Exception, match='UNIQUE constraint failed'):
        await txn.exq(query)


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_insert_or_fail_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().insert_or_fail(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == (
        'INSERT OR FAIL INTO "tbl" ("name", "age") VALUES(?, ?)',
        ('Ringo', 43),
    )
    with pytest.raises(Exception, match='UNIQUE constraint failed'):
        await txn.exq(query)


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        'mysql',
    ],
    indirect=True,
)
async def test_insert_or_ignore_into(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = Query().insert_or_ignore(tbl).columns(tbl.name, tbl.age).values(('Ringo', 43))
    assert query.__getsql__() == (
        'INSERT OR IGNORE INTO "tbl" ("name", "age") VALUES(?, ?)',
        ('Ringo', 43),
    )
    assert await txn.exq(query) == []


async def test_insert_from_select(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    query = (
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .select(
            Query().select(
                ExprLiteral('George').typed(str),
                ExprLiteral(42).typed(int),
            ),
        )
    )
    assert query.__getsql__() == ('INSERT INTO "tbl" ("name", "age") SELECT ?, ?', ('George', 42))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_insert_on_conflict_do_nothing(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    query = (
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(('Ringo', 43))
        .on_conflict_do_nothing()
    )
    assert query.__getsql__() == (
        'INSERT INTO "tbl" ("name", "age") VALUES(?, ?) ON CONFLICT DO NOTHING',
        ('Ringo', 43),
    )
    assert await txn.exq(query) == []
    assert await txn.exq(Query().select(tbl.age.typed(int)).from_(tbl)) == [(42,)]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_insert_on_conflict_update_set(txn: Transaction) -> None:
    """Test insert statement."""
    tbl = Table('tbl')
    excluded = Table('excluded')
    await txn.exq(Query().insert(tbl).columns(tbl.name, tbl.age).values(('Ringo', 42)))

    with pytest.raises(QueryError, match='same length'):
        (
            Query()
            .insert(tbl)
            .columns(tbl.name, tbl.age)
            .values(('Ringo', 40))
            .on_conflict(tbl.name)
            .update_set((tbl.name, tbl.age), (excluded.age,))
            .where(excluded.age > tbl.age)
            .__getsql__()
        )

    query = (
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(('Ringo', 40))
        .on_conflict(tbl.name)
        .update_set((tbl.age,), (excluded.age,))
        .where(excluded.age > tbl.age)
    )
    assert query.__getsql__() == (
        'INSERT INTO "tbl" ("name", "age") VALUES(?, ?) ON CONFLICT ("tbl"."name") DO '
        'UPDATE SET age = "excluded"."age" WHERE "excluded"."age" > "tbl"."age"',
        ('Ringo', 40),
    )
    assert await txn.exq(query) == []
    assert await txn.exq(Query().select(tbl.age.typed(int)).from_(tbl)) == [(42,)]

    query = (
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(('Ringo', 43))
        .on_conflict(tbl.name)
        .update_set((tbl.age,), (excluded.age,))
        .where(excluded.age > tbl.age)
    )
    assert query.__getsql__() == (
        'INSERT INTO "tbl" ("name", "age") VALUES(?, ?) ON CONFLICT ("tbl"."name") DO '
        'UPDATE SET age = "excluded"."age" WHERE "excluded"."age" > "tbl"."age"',
        ('Ringo', 43),
    )
    assert await txn.exq(query) == []
    assert await txn.exq(Query().select(tbl.age.typed(int)).from_(tbl)) == [(43,)]

    query = (
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(('Ringo', 44))
        .on_conflict(tbl.name)
        .update_set((tbl.age,), (0,))
        .where(excluded.age > tbl.age)
    )
    assert query.__getsql__() == (
        'INSERT INTO "tbl" ("name", "age") VALUES(?, ?) ON CONFLICT ("tbl"."name") DO '
        'UPDATE SET age = ? WHERE "excluded"."age" > "tbl"."age"',
        ('Ringo', 44, 0),
    )
    assert await txn.exq(query) == []
    assert await txn.exq(Query().select(tbl.age.typed(int)).from_(tbl)) == [(0,)]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Models declarative layer tests."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

from sqls.models import (
    CharField,
    Fieldmeta,
    ForeignKeyField,
    IntegerField,
    Model,
    Tablemeta,
    get_create_queries,
)
from sqls.models.model import Indexmeta


def get_q(*models: type[Model]) -> list[str]:
    """Convert models to create statements."""
    queries = get_create_queries(models)
    return [x.__getsql__()[0] for x in queries]


def test_simple_declaration() -> None:
    """Test simple field."""

    @dataclass
    class Node(Model):
        """Test model."""

        num: IntegerField

    assert get_q(Node) == [
        (
            'CREATE TABLE IF NOT EXISTS "node" '
            '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "num" INTEGER NOT NULL)'
        ),
    ]


def test_annotated_declaration() -> None:
    """Test annotated field."""

    @dataclass
    class Node(Model):
        """Test model."""

        text: Annotated[CharField, Fieldmeta(max_length=32)]

    assert get_q(Node) == [
        (
            'CREATE TABLE IF NOT EXISTS "node" '
            '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "text" VARCHAR(32) NOT NULL)'
        ),
    ]


def test_unique_fields() -> None:
    """Test unique fields."""

    @dataclass
    class Node(Model):
        """Test model."""

        optional: IntegerField | None
        unique: Annotated[IntegerField, Fieldmeta(unique=True)]
        unique_optional: Annotated[IntegerField, Fieldmeta(unique=True)] | None

    assert get_q(Node) == [
        (
            'CREATE TABLE IF NOT EXISTS "node" '
            '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "optional" INTEGER, "unique" '
            'INTEGER NOT NULL UNIQUE, "unique_optional" INTEGER UNIQUE)'
        ),
    ]


def test_default_value() -> None:
    """Test unique fields."""

    @dataclass
    class Node(Model):
        """Test model."""

        defaulted: Annotated[IntegerField, Fieldmeta(default=42)]

    assert get_q(Node) == [
        (
            'CREATE TABLE IF NOT EXISTS "node" '
            '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,'
            ' "defaulted" INTEGER NOT NULL DEFAULT 42)'
        ),
    ]


def test_foreignkey_selfreference() -> None:
    """Test self references."""

    @dataclass
    class Node(Model):
        """Test model."""

        node_id: ForeignKeyField
        parent: Annotated[ForeignKeyField, Fieldmeta(foreign_key=('node', 'id'))]

    assert get_q(Node) == [
        'CREATE TABLE IF NOT EXISTS "node" '
        '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,'
        ' "node_id" INTEGER NOT NULL REFERENCES "node" ("id") ON DELETE CASCADE,'
        ' "parent" INTEGER NOT NULL REFERENCES "node" ("id") ON DELETE CASCADE)',
        'CREATE INDEX IF NOT EXISTS "idx__node__node_id" ON "node" ("node_id")',
        'CREATE INDEX IF NOT EXISTS "idx__node__parent" ON "node" ("parent")',
    ]
    assert Node.__model__['backward_rels'] == [(Node, 'node_id', 'id'), (Node, 'parent', 'id')]
    assert Node.__model__['forward_rels'] == [(Node, 'id', 'node_id'), (Node, 'id', 'parent')]


def test_foreignkey_reference() -> None:
    """Test references."""

    @dataclass
    class Parent(Model):
        """Test model."""

    @dataclass
    class Child(Model):
        """Test model."""

        parent_id: ForeignKeyField

    assert get_q(Parent, Child) == [
        'CREATE TABLE IF NOT EXISTS "parent" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL)',
        'CREATE TABLE IF NOT EXISTS "child" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
        '"parent_id" INTEGER NOT NULL REFERENCES "parent" ("id") ON DELETE CASCADE)',
        'CREATE INDEX IF NOT EXISTS "idx__child__parent_id" ON "child" ("parent_id")',
    ]
    assert Parent.__model__['backward_rels'] == [(Child, 'parent_id', 'id')]
    assert Child.__model__['forward_rels'] == [(Parent, 'id', 'parent_id')]


def test_foreignkey_reference_with_restrict() -> None:
    """Test references."""

    @dataclass
    class Parent(Model):
        """Test model."""

    @dataclass
    class Child(Model):
        """Test model."""

        parent_id: Annotated[ForeignKeyField, Fieldmeta(on_delete='RESTRICT')]

    assert get_q(Parent, Child) == [
        'CREATE TABLE IF NOT EXISTS "parent" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL)',
        'CREATE TABLE IF NOT EXISTS "child" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
        '"parent_id" INTEGER NOT NULL REFERENCES "parent" ("id") ON DELETE RESTRICT)',
        'CREATE INDEX IF NOT EXISTS "idx__child__parent_id" ON "child" ("parent_id")',
    ]
    assert Parent.__model__['backward_rels'] == [(Child, 'parent_id', 'id')]
    assert Child.__model__['forward_rels'] == [(Parent, 'id', 'parent_id')]


def test_foreignkey_reference_with_idfield() -> None:
    """Test self references."""

    @dataclass
    class Node(Model):
        """Test model."""

        id: Annotated[ForeignKeyField, Fieldmeta(foreign_key=('node', 'id'))]

    assert get_q(Node) == [
        'CREATE TABLE IF NOT EXISTS "node" '
        '("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL'
        ' REFERENCES "node" ("id") ON DELETE CASCADE)',
    ]
    assert Node.__model__['backward_rels'] == [(Node, 'id', 'id')]
    assert Node.__model__['forward_rels'] == [(Node, 'id', 'id')]


def test_through_reference() -> None:
    """Test references."""

    @dataclass
    class User(Model):
        """Test model."""

    @dataclass
    class Group(Model):
        """Test model."""

    @dataclass
    class UserGroup(Model):
        """Test model."""

        id: None
        user_id: ForeignKeyField
        group_id: ForeignKeyField

    assert get_q(User, Group, UserGroup) == [
        'CREATE TABLE IF NOT EXISTS "user" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL)',
        'CREATE TABLE IF NOT EXISTS "group" ("id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL)',
        'CREATE TABLE IF NOT EXISTS "user_group" ('
        '"user_id" INTEGER NOT NULL REFERENCES "user" ("id") ON DELETE CASCADE, '
        '"group_id" INTEGER NOT NULL REFERENCES "group" ("id") ON DELETE CASCADE)',
        'CREATE INDEX IF NOT EXISTS "idx__user_group__user_id" ON "user_group" ("user_id")',
        'CREATE INDEX IF NOT EXISTS "idx__user_group__group_id" ON "user_group" ("group_id")',
    ]
    assert User.__model__['through_rels'] == [(UserGroup, 'user_id', 'id', Group, 'id', 'group_id')]
    assert Group.__model__['through_rels'] == [(UserGroup, 'group_id', 'id', User, 'id', 'user_id')]
    assert not UserGroup.__model__['forward_rels']
    assert not UserGroup.__model__['backward_rels']


def test_compound_index() -> None:
    """Test compound index."""

    @dataclass
    class UserGroup(Model):
        """Test model."""

        __tablemeta__ = Tablemeta(
            indexes=(Indexmeta(fields=('user_id', 'group_id'), unique=False),),
        )

        id: None
        user_id: IntegerField
        group_id: IntegerField

    assert get_q(UserGroup) == [
        'CREATE TABLE IF NOT EXISTS "user_group" ('
        '"user_id" INTEGER NOT NULL, '
        '"group_id" INTEGER NOT NULL)',
        'CREATE INDEX IF NOT EXISTS "idx__user_group__user_id__group_id" ON '
        '"user_group" ("user_id", "group_id")',
    ]


def test_compound_unique_index() -> None:
    """Test compound unique index."""

    @dataclass
    class UserGroup(Model):
        """Test model."""

        __tablemeta__ = Tablemeta(indexes=(Indexmeta(fields=('user_id', 'group_id'), unique=True),))

        id: None
        user_id: IntegerField
        group_id: IntegerField

    assert get_q(UserGroup) == [
        'CREATE TABLE IF NOT EXISTS "user_group" ('
        '"user_id" INTEGER NOT NULL, '
        '"group_id" INTEGER NOT NULL)',
        'CREATE UNIQUE INDEX IF NOT EXISTS "idx__user_group__user_id__group_id" ON '
        '"user_group" ("user_id", "group_id")',
    ]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""CREATE TRIGGER tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key()
        .column('name', 'TEXT')
        .column('other', 'TEXT'),
    )

    dst = Table('dst')
    await txn.exq(
        Query()
        .create_table(dst)
        .column('id', 'INTEGER')
        .primary_key()
        .column('cause', 'TEXT')
        .column('columns', 'TEXT'),
    )

    view = Table('view')
    await txn.exq(
        Query().create_view(view).as_(Query().select(tbl.__star__.typed(object)).from_(tbl)),
    )


async def get_lastentry(txn: Transaction) -> tuple[int, str, str]:
    """Get last entry from dst."""
    dst = Table('dst')
    return (
        await txn.exq(
            Query()
            .select(dst.id.typed(int), dst.cause.typed(str), dst.columns.typed(str))
            .from_(dst)
            .order_by(dst.id, desc=True)
            .limit(1),
        )
    )[0]


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_after_insert(txn: Transaction) -> None:
    """Test create trigger statement."""
    tbl = Table('tbl')
    dst = Table('dst')

    new = Table('NEW')
    assert (
        await txn.exq(
            Query()
            .create_trigger(Table('after_insert'))
            .after()
            .insert()
            .on_(tbl)
            .for_each_row()
            .when(new.name != ExprLiteral('ignored', bind=False))
            .statements(
                Query()
                .insert(dst)
                .columns(dst.cause, dst.columns)
                .values(ExprLiteral(('tbl_insert', new.name), bind=False)),
            ),
        )
        == []
    )

    assert await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',))) == []
    assert await get_lastentry(txn) == (1, 'tbl_insert', 'foo')
    assert await txn.exq(Query().insert(tbl).columns(tbl.name).values(('ignored',))) == []
    assert await get_lastentry(txn) == (1, 'tbl_insert', 'foo')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_before_update(txn: Transaction) -> None:
    """Test create trigger statement."""
    tbl = Table('tbl')
    dst = Table('dst')

    old = Table('OLD')
    new = Table('NEW')
    assert (
        await txn.exq(
            Query()
            .create_trigger(Table('before_update'))
            .before()
            .update()
            .on_(tbl)
            .for_each_row()
            .when(new.name != old.name)
            .statements(
                Query()
                .insert(dst)
                .columns(dst.cause, dst.columns)
                .values(
                    (
                        ExprLiteral('tbl_update', bind=False),
                        old.name.strconcat(ExprLiteral('->', bind=False)).strconcat(new.name),
                    ),
                ),
            ),
        )
        == []
    )
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',)))
    await txn.exq(Query().update(tbl).set((tbl.name,), ('bar',)).where(tbl.name == 'foo'))
    assert await get_lastentry(txn) == (1, 'tbl_update', 'foo->bar')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_before_update_of(txn: Transaction) -> None:
    """Test create trigger statement."""
    tbl = Table('tbl')
    dst = Table('dst')

    new = Table('NEW')
    assert (
        await txn.exq(
            Query()
            .create_trigger(Table('before_update_of'))
            .before()
            .update(of_=[tbl.other])
            .on_(tbl)
            .for_each_row()
            .statements(
                Query()
                .insert(dst)
                .columns(dst.cause, dst.columns)
                .values(
                    (
                        ExprLiteral('tbl_update_other', bind=False),
                        ExprLiteral('->', bind=False).strconcat(new.other),
                    ),
                ),
            ),
        )
        == []
    )
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',)))
    await txn.exq(Query().update(tbl).set((tbl.other,), ('bar',)).where(tbl.name == 'foo'))
    assert await get_lastentry(txn) == (1, 'tbl_update_other', '->bar')
    await txn.exq(Query().update(tbl).set((tbl.name,), ('bar',)).where(tbl.name == 'foo'))
    assert await get_lastentry(txn) == (1, 'tbl_update_other', '->bar')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_delete(txn: Transaction) -> None:
    """Test create trigger statement."""
    tbl = Table('tbl')
    dst = Table('dst')

    old = Table('OLD')
    assert (
        await txn.exq(
            Query()
            .create_trigger(Table('delete'))
            .delete()
            .on_(tbl)
            .statements(
                Query()
                .insert(dst)
                .columns(dst.cause, dst.columns)
                .values((ExprLiteral('tbl_delete', bind=False), old.name)),
            ),
        )
        == []
    )
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',)))
    await txn.exq(Query().delete(tbl).where(tbl.name == 'foo'))
    assert await get_lastentry(txn) == (1, 'tbl_delete', 'foo')


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='other syntax')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='other syntax')),
    ],
    indirect=True,
)
async def test_instead_of(txn: Transaction) -> None:
    """Test create trigger statement."""
    dst = Table('dst')
    view = Table('view')

    new = Table('NEW')
    assert (
        await txn.exq(
            Query()
            .create_trigger(Table('instead_of'))
            .instead_of()
            .insert()
            .on_(view)
            .statements(
                Query()
                .insert(dst)
                .columns(dst.cause, dst.columns)
                .values(
                    (
                        ExprLiteral('instead_of', bind=False),
                        new.name.strconcat(ExprLiteral(' ', bind=False).strconcat(new.other)),
                    ),
                ),
            ),
        )
        == []
    )
    await txn.exq(Query().insert(view).columns(view.name, view.other).values(('vname', 'vother')))
    assert await get_lastentry(txn) == (1, 'instead_of', 'vname vother')



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL CREATE TABLE."""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from .base import QBase, Term

if TYPE_CHECKING:
    from sqls.interfaces import SqlValue

    from .expr import Expr, ExprColumn
    from .table import Table

T1 = TypeVar('T1', bound='QBase[tuple[SqlValue, ...]]')
T2 = TypeVar('T2', bound='QBase[tuple[SqlValue, ...]]')
T3 = TypeVar('T3', bound='QBase[tuple[SqlValue, ...]]')


class CreateTable(QBase[tuple[()]]):
    """CREATE TABLE mixin."""

    def create_table(
        self,
        table: Table,
        temp: bool = False,
        if_not_exists: bool = False,
    ) -> PostCreate:
        """Start CREATE TABLE statement.

        Args:
            table: New table.
            temp: Should table be temporary.
            if_not_exists: Ignore if table exists.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()
        tmp = ' TEMP' if temp else ''
        ine = ' IF NOT EXISTS' if if_not_exists else ''
        return self._f(PostCreate, f'CREATE{tmp} TABLE{ine} {sql}', ())


class ColumnDef(QBase[tuple[()]]):
    """Column definition mixin."""

    def column(self, name: str, typ: str | None = None) -> PostColumnDef:
        """Add column definition.

        Args:
            name: Column name.
            typ: Column type.

        Returns:
            Query builder.

        """
        if self._pis(CreateTable):
            paren = ' ('
            enter = ((WithoutRowid, Term), ')')
        else:
            paren = ', '
            enter = None
        typ = f' {typ}' if typ else ''
        return self._f(PostColumnDef, f'{paren}"{name}"{typ}', enter=enter)


class ColumnConstraintGeneric(QBase[tuple[()]], Generic[T1, T2, T3]):
    """Generic column constraint mixin."""

    _next_columnconstraint: type[T1]
    _next_columnconstraint_pre: type[T2]
    _next_columnconstraint_references: type[T3]

    def primary_key(
        self,
        sort: str | None = None,
        autoincrement: bool | None = None,
    ) -> T1:
        """Add PRIMARY KEY.

        Args:
            sort: Sort order.
            autoincrement: Should add autoincrement.

        Returns:
            Query builder.

        """
        if autoincrement:
            enter = (
                (
                    ColumnDef,
                    ColumnConstraintGeneric,
                    TableConstraint,
                    WithoutRowid,
                    Term,
                ),
                ' AUTOINCREMENT',
            )
        else:
            enter = None
        sort = f' {sort}' if sort else ''
        return self._f(
            self._next_columnconstraint,
            f' PRIMARY KEY{sort}',
            enter=enter,
        )

    def not_null(self) -> T1:
        """Add NOT NULL constraint."""
        return self._f(self._next_columnconstraint, ' NOT NULL')

    def unique(self) -> T1:
        """Add UNIQUE constraint."""
        return self._f(self._next_columnconstraint, ' UNIQUE')

    def check(self, expr: Expr) -> T2:
        """Add CHECK constraint.

        Args:
            expr: Expression to check.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(self._next_columnconstraint_pre, f' CHECK ({sql})', args)

    def default(self, value: SqlValue) -> T2:
        """Add DEFAULT clause.

        Args:
            value: Default value.

        Returns:
            Query builder.

        """
        if value is None:
            return self._f(self._next_columnconstraint_pre, ' DEFAULT NULL')
        if value is True:
            return self._f(self._next_columnconstraint_pre, ' DEFAULT TRUE')
        if value is False:
            return self._f(self._next_columnconstraint_pre, ' DEFAULT FALSE')
        return self._f(self._next_columnconstraint_pre, f' DEFAULT {value!r}')

    def collate(self, name: str) -> T2:
        """Add COLLATE clause.

        Args:
            name: Collation name.

        Returns:
            Query builder.

        """
        return self._f(self._next_columnconstraint_pre, f' COLLATE {name}')

    def references(self, table: Table, *cols: ExprColumn) -> T3:
        """Add foreign-key clause.

        Args:
            table: Referenced table.
            cols: Referenced columns.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()

        def colname(col: ExprColumn) -> str:
            return f'"{col.col}"'

        return self._f(
            self._next_columnconstraint_references,
            f' REFERENCES {sql} ({", ".join(colname(x) for x in cols)})',
        )

    def generated(self, expr: Expr, stored: bool = False) -> T2:
        """Add GENERATED clause.

        Args:
            expr: Expression.
            stored: Use STORED instead of VIRTUAL.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        typ = 'STORED' if stored else 'VIRTUAL'
        return self._f(self._next_columnconstraint_pre, f' GENERATED ALWAYS AS ({sql}) {typ}', args)


class ColumnConstraint(  # yapf: prevent format
    ColumnConstraintGeneric[
        'PostColumnConstraint',
        'PostColumnDef',
        'PostColumnConstraintReferences',
    ],
):
    """Column constraint mixin."""


class OnRefchangeGeneric(QBase[tuple[()]], Generic[T1]):
    """Generic on reference change mixin."""

    _next_onrefchange: type[T1]

    def on_delete(self) -> T1:
        """Add ON DELETE action."""
        return self._f(self._next_onrefchange, ' ON DELETE')

    def on_update(self) -> T1:
        """Add ON UPDATE action."""
        return self._f(self._next_onrefchange, ' ON UPDATE')


class OnRefchange(OnRefchangeGeneric['OnRefchangeAction']):
    """On reference change mixin."""


class OnRefchangeActionGeneric(QBase[tuple[()]], Generic[T1]):
    """Generic on reference change action mixin."""

    _next_onrefchangeaction: type[T1]

    def set_null(self) -> T1:
        """Add SET NULL action."""
        return self._f(self._next_onrefchangeaction, ' SET NULL')

    def set_default(self) -> T1:
        """Add SET DEFAULT action."""
        return self._f(self._next_onrefchangeaction, ' SET DEFAULT')

    def cascade(self) -> T1:
        """Add CASCASE action."""
        return self._f(self._next_onrefchangeaction, ' CASCADE')

    def restrict(self) -> T1:
        """Add CASCASE action."""
        return self._f(self._next_onrefchangeaction, ' RESTRICT')

    def no_action(self) -> T1:
        """Add NO ACTION."""
        return self._f(self._next_onrefchangeaction, ' NO ACTION')


class OnRefchangeAction(OnRefchangeActionGeneric['PostColumnConstraintReferences']):
    """On reference change action mixin."""


class DeferrableGeneric(QBase[tuple[()]], Generic[T1]):
    """Generic deferrable mixin."""

    _next_deferrable: type[T1]

    def deferrable(self, is_deferrable: bool = True, initially: str | None = None) -> T1:
        """Add DEFERRABLE clause.

        Args:
            is_deferrable: Is column deferrable.
            initially: Initial state.

        Returns:
            Query builder.

        """
        neg = ' NOT' if not is_deferrable else ''
        initial = f' INITIALLY {initially}' if initially else ''
        return self._f(self._next_deferrable, f'{neg} DEFERRABLE{initial}')


class Deferrable(DeferrableGeneric['PostColumnDef']):
    """Deferrable mixin."""


class TableConstraint(QBase[tuple[()]]):
    """Table constraint mixin."""

    def unique_columns(self, *names: str) -> PostColumns:
        """Add UNIQUE table constraint.

        Args:
            names: Column names.

        Returns:
            Query builder.

        """
        return self._f(PostColumns, f', UNIQUE ({", ".join(names)})')

    def check_columns(self, expr: Expr) -> PostColumns:
        """Add CHECK table constraint.

        Args:
            expr: Expression.

        Returns:
            Query builder.

        """
        sql, args = expr.__getsql__()
        return self._f(PostColumns, f', CHECK ({sql})', args)

    def foreign_key(self, *names: str) -> ForeignKey:
        """Start foreign-key table constraint.

        Args:
            names: Column names.

        Returns:
            Query builder.

        """
        return self._f(ForeignKey, f', FOREIGN KEY ({", ".join(names)})')


class ForeignKey(QBase[tuple[()]]):
    """Foreign key mixin."""

    def references(self, table: Table, *cols: ExprColumn) -> PostTableConstraint:
        """Add foreign-key references.

        Args:
            table: Referenced table.
            cols: Referenced columns.

        Returns:
            Query builder.

        """
        sql, _ = table.__getsql__()

        def colname(col: ExprColumn) -> str:
            return f'"{col.col}"'

        return self._f(
            PostTableConstraint,
            f' REFERENCES {sql} ({", ".join(colname(x) for x in cols)})',
        )


class OnConflictGeneric(QBase[tuple[()]], Generic[T1]):
    """Generic on conflict mixin."""

    _next_conflictclause: type[T1]

    def _on_conflict(self, term: str) -> T1:
        """Add ON CONFLICT clause.

        Args:
            term: Action to take.

        Returns:
            Query builder.

        """
        return self._f(self._next_conflictclause, f' ON CONFLICT {term}')

    def on_conflict_rollback(self) -> T1:
        """Add ON CONFLICT ROLLBACK."""
        return self._on_conflict('ROLLBACK')

    def on_conflict_abort(self) -> T1:
        """Add ON CONFLICT ABORT."""
        return self._on_conflict('ABORT')

    def on_conflict_fail(self) -> T1:
        """Add ON CONFLICT FAIL."""
        return self._on_conflict('FAIL')

    def on_conflict_ignore(self) -> T1:
        """Add ON CONFLICT IGNORE."""
        return self._on_conflict('IGNORE')

    def on_conflict_replace(self) -> T1:
        """Add ON CONFLICT REPLACE."""
        return self._on_conflict('REPLACE')


class OnConflict(OnConflictGeneric['PostConflict']):
    """On conflict mixin."""


class OnRefchangeTbl(OnRefchangeGeneric['OnRefchangeTblAction']):
    """On reference change table mixin."""


class OnRefchangeTblAction(OnRefchangeActionGeneric['PostTableConstraint']):
    """On reference change action table mixing."""


class DeferrableTbl(DeferrableGeneric['PostColumns']):
    """Deferrable table mixin."""


class WithoutRowid(QBase[tuple[()]]):
    """WITHOUT ROWID mixin."""

    def without_rowid(self) -> Term[tuple[()]]:
        """Enable WITHOUT ROWID."""
        return self._f(Term, ' WITHOUT ROWID')


class PostCreate(
    ColumnDef,
):
    """Post create builder."""


class PostColumnDef(
    ColumnDef,
    ColumnConstraint,
    TableConstraint,
    WithoutRowid,
    Term[tuple[()]],
):
    """Post column definition builder."""


class PostColumnConstraint(
    PostColumnDef,
    OnConflict,
):
    """Post column constraint builder."""


class PostColumnConstraintReferences(
    PostColumnDef,
    OnRefchange,
    Deferrable,
):
    """Post column constraint references builder."""


class PostConflict(
    PostColumnDef,
):
    """Post conflict builder."""


class PostColumns(
    TableConstraint,
    WithoutRowid,
    Term[tuple[()]],
):
    """Post columns builder."""


class PostTableConstraint(
    PostColumns,
    OnRefchangeTbl,
    DeferrableTbl,
):
    """Post table constraint builder."""



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL Query."""

from .alter import Alter
from .analyze import Analyze
from .attach import Attach
from .create_index import CreateIndex
from .create_table import CreateTable
from .create_trigger import CreateTrigger
from .create_view import CreateView
from .delete import Delete
from .detach import Detach
from .drop import Drop
from .insert import Insert
from .pragma import Pragma
from .reindex import Reindex
from .select import Select
from .update import Update
from .vacuum import Vacuum
from .withq import With


class Query(
    Alter,
    Analyze,
    Attach,
    CreateIndex,
    CreateTable,
    CreateTrigger,
    CreateView,
    Delete,
    Detach,
    Drop,
    Insert,
    Pragma,
    Reindex,
    Select,
    Update,
    Vacuum,
    With,
):
    """SQL Query Builder.

    All regular SQL queries are build up from here.

    """



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""UPDATE tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, QueryError, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key(autoincrement=True)
        .unique()
        .column('name', 'TEXT')
        .unique()
        .column('age', 'INTEGER'),
    )

    dst = Table('dst')
    await txn.exq(
        Query()
        .create_table(dst)
        .column('id', 'INTEGER')
        .primary_key()
        .column('name', 'TEXT')
        .column('age', 'INTEGER'),
    )

    await txn.exq(
        Query()
        .insert(tbl)
        .columns(tbl.name, tbl.age)
        .values(
            ('John', 42),
            ('Paul', 43),
            ('Ringo', 44),
            ('George', 45),
        ),
    )


async def test_column_all(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    with pytest.raises(QueryError, match='same length'):
        Query().update(tbl).set((tbl.age,), ())

    query = Query().update(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_update_or_rollback(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update_or_rollback(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE OR ROLLBACK "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_update_or_abort(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update_or_abort(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE OR ABORT "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_update_or_replace(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update_or_replace(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE OR REPLACE "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_update_or_fail(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update_or_fail(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE OR FAIL "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        'mysql',
    ],
    indirect=True,
)
async def test_update_or_ignore(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update_or_ignore(tbl).set((tbl.age,), (42,))
    assert query.__getsql__() == ('UPDATE OR IGNORE "tbl" SET "age"=?', (42,))
    assert await txn.exq(query) == []


async def test_where(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update(tbl).set((tbl.age,), (42,)).where(tbl.name == 'foo')
    assert query.__getsql__() == (
        'UPDATE "tbl" SET "age"=? WHERE "tbl"."name" = ?',
        (42, 'foo'),
    )
    assert await txn.exq(query) == []


async def test_expr(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    query = Query().update(tbl).set((tbl.age,), (tbl.age + 1,)).where(tbl.name == 'foo')
    assert query.__getsql__() == (
        'UPDATE "tbl" SET "age"="tbl"."age" + ? WHERE "tbl"."name" = ?',
        (1, 'foo'),
    )
    assert await txn.exq(query) == []


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        'postgres',
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_from(txn: Transaction) -> None:
    """Test update statement."""
    tbl = Table('tbl')
    dst = Table('dst')

    query = Query().update(tbl).set((tbl.age,), (dst.age,)).from_(dst)
    assert query.__getsql__() == ('UPDATE "tbl" SET "age"="dst"."age" FROM "dst"', ())
    assert await txn.exq(query) == []

    other_query = (
        Query().update(tbl).set((tbl.age,), (dst.age,)).from_(dst).where(tbl.name == dst.name)
    )
    assert other_query.__getsql__() == (
        'UPDATE "tbl" SET "age"="dst"."age" FROM "dst" WHERE "tbl"."name" = "dst"."name"',
        (),
    )
    assert await txn.exq(other_query) == []



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""CREATE VIEW tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.fixture(autouse=True)
async def _autotable(txn: Transaction) -> None:
    tbl = Table('tbl')
    await txn.exq(
        Query()
        .create_table(tbl)
        .column('id', 'INTEGER')
        .primary_key(autoincrement=True)
        .column('name', 'TEXT'),
    )


async def test_simple(txn: Transaction) -> None:
    """Test create view statement."""
    tbl = Table('tbl')
    view = Table('vw_')

    await txn.exq(
        Query()
        .create_view(view)
        .as_(
            Query().select(tbl.id.typed(int), tbl.name.typed(str)).from_(tbl),
        ),
    )
    assert not await txn.exq(Query().select(view.id.typed(int), view.name.typed(str)).from_(view))
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',), ('bar',)))
    assert await txn.exq(Query().select(view.id.typed(int), view.name.typed(str)).from_(view)) == [
        (1, 'foo'),
        (2, 'bar'),
    ]


async def test_column_names(txn: Transaction) -> None:
    """Test create view statement."""
    tbl = Table('tbl')
    view = Table('vw_')

    await txn.exq(
        Query()
        .create_view(view)
        .columns(view.name, view.id)
        .as_(
            Query().select(tbl.name.typed(int), tbl.id.typed(str)).from_(tbl),
        ),
    )
    assert not await txn.exq(Query().select(view.id.typed(int), view.name.typed(str)).from_(view))
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',), ('bar',)))
    assert await txn.exq(Query().select(view.id.typed(int), view.name.typed(str)).from_(view)) == [
        (1, 'foo'),
        (2, 'bar'),
    ]



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Squal tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sqls.queries import ExprLiteral, Query, Table

if TYPE_CHECKING:
    from sqls.interfaces import Transaction


@pytest.mark.parametrize(
    'txn',
    [
        'sqlite',
        pytest.param('postgres', marks=pytest.mark.skip(reason='unsupported')),
        pytest.param('mysql', marks=pytest.mark.skip(reason='unsupported')),
    ],
    indirect=True,
)
async def test_attach_detach(txn: Transaction) -> None:
    """Test attach/detach statement."""
    tbl = Table('tbl')
    atbl = Table('atbl', schema='attached')
    await txn.exq(
        Query().create_table(tbl).column('id', 'INTEGER').primary_key().column('name', 'TEXT'),
    )
    await txn.exq(Query().insert(tbl).columns(tbl.name).values(('foo',)))

    query = Query().attach(ExprLiteral(':memory:'), 'attached')
    assert await txn.exq(query) == []

    query = Query().create_table(atbl).column('id', 'INTEGER').primary_key().column('name', 'TEXT')
    assert await txn.exq(query) == []

    query = Query().insert(atbl).columns(atbl.name).values(('bar',))
    assert await txn.exq(query) == []

    assert await txn.exq(Query().select(atbl.name.typed(str)).from_(atbl)) == [('bar',)]

    await txn.execute('COMMIT')
    await txn.execute('BEGIN')

    query = Query().detach('attached')
    assert await txn.exq(query) == []

    with pytest.raises(Exception, match='no such table') as exc:
        assert await txn.exq(Query().select(atbl.name.typed(str)).from_(atbl))
    assert 'no such table' in str(exc)



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""SQL ANALYZE."""

from .base import QBase, Term
from .table import Table


class Analyze(QBase[tuple[()]]):
    """ANALYZE mixin."""

    def analyze(self, target: str | Table | None = None) -> Term[tuple[()]]:
        """Analyze target."""
        if target:
            if not isinstance(target, Table):
                target = Table(target)
            sql, _ = target.__getsql__()
            sql = f' {sql}'
        else:
            sql = ''
        return self._f(Term, f'ANALYZE{sql}')



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Model base class and utilities."""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, NamedTuple, TypedDict

from .fields import AutoField  # noqa: TCH001

if TYPE_CHECKING:
    from typing import TypeVar

    from sqls.interfaces import SqlValue

    T = TypeVar('T')

RE_SNAKE = re.compile(r'(?<!^)(?=[A-Z])')


class Indexmeta(NamedTuple):
    """Index metadata."""

    fields: tuple[str, ...]
    unique: bool = False


class Tablemeta(NamedTuple):
    """Table configuration."""

    schema: str | None = None
    name: str | None = None
    indexes: tuple[Indexmeta, ...] = ()


class Fieldmeta(NamedTuple):
    """Field configuration."""

    name: str | None = None
    primary: bool = False
    unique: bool = False
    collation: str | None = None
    default: SqlValue = None
    foreign_key: tuple[str, str] | None = None
    max_length: int | None = None
    on_delete: str | None = None


class Metadata(TypedDict):
    """Model metadata."""

    name: str
    temp_rels: list[tuple[bool, str, str, str]]
    forward_rels: list[tuple[type[Model], str, str]]
    backward_rels: list[tuple[type[Model], str, str]]
    through_rels: list[tuple[type[Model], str, str, type[Model], str, str]]


@dataclass
class Model:
    """Model mixin."""

    __tablemeta__: ClassVar[Tablemeta] = Tablemeta()
    __model__: ClassVar[Metadata]  # = {}

    id: AutoField | None

    def __init_subclass__(cls) -> None:
        """Initialize __model__ attribute."""
        cls.__model__ = {
            'name': RE_SNAKE.sub('_', cls.__name__).lower(),
            'temp_rels': [],
            'forward_rels': [],
            'backward_rels': [],
            'through_rels': [],
        }



# Copyright 2020 - 2024 Marko Durkovic
# SPDX-License-Identifier: Apache-2.0
"""Query base."""

from __future__ import annotations

import sys
from collections.abc import Callable  # noqa: TCH003
from functools import wraps
from typing import (
    TYPE_CHECKING,
    ClassVar,
    ForwardRef,
    Generic,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from sqls.interfaces import BindArg, SqlValue

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Concatenate, ParamSpec, TypeAlias

    P = ParamSpec('P')
    R = TypeVar('R')

    Other = TypeVar('Other', bound='QBase[Texplicit]')

Texplicit: TypeAlias = tuple[SqlValue | object, ...]
T_co = TypeVar('T_co', bound='Texplicit', covariant=True)


def get_generics(cls: type[QBase[T_co]]) -> Generator[type[QBase[Texplicit]], None, None]:
    """Get generic parents."""
    for parent in getattr(cls, '__orig_bases__', []):
        origin = get_origin(parent)
        if not origin or not issubclass(origin, QBase):
            continue
        yield parent
        yield from get_generics(origin)


class QBase(Generic[T_co]):
    """Base for queries."""

    _unfinished: ClassVar[dict[tuple[str, str], list[type[QBase[Texplicit]]]]] = {}

    _parts: list[str]
    _args: list[BindArg]
    _enterstack: list[tuple[tuple[type[QBase[Texplicit]], ...], str]]

    endq: Callable[[], Term[T_co]]

    attr: T_co | None = None

    def __init__(self, parent: QBase[Texplicit] | None = None) -> None:
        """Initialize."""
        super().__init__()
        self._parent: QBase[Texplicit] | None
        if parent:
            self._parts = parent._parts[:]  # noqa: SLF001
            self._args = parent._args[:]  # noqa: SLF001
            self._enterstack = parent._enterstack[:]  # noqa: SLF001
            self._parent = parent
        else:
            self._parts = []
            self._args = []
            self._enterstack = []
            self._parent = None
        self._calledon: type | None = None

    def _append(self, other: QBase[Texplicit]) -> None:
        """Forward own state to other."""
        assert not self._enterstack
        self._parts += other._parts  # noqa: SLF001
        self._args += other._args  # noqa: SLF001

    def _pis(self, *classes: type) -> bool:
        """Check if parent is instance of class."""
        assert self._parent
        return self._parent._calledon in classes  # noqa: SLF001

    def _f(
        self,
        cls: type[Other],
        parts: str,
        args: tuple[BindArg, ...] = (),
        enter: tuple[tuple[type[QBase[Texplicit]], ...], str] | None = None,
    ) -> Other:
        """Forward self and args to next."""
        inst = cls(cast('QBase[tuple[SqlValue, ...]]', self))
        inst._parts += [parts]  # noqa: SLF001
        inst._args += args  # noqa: SLF001
        if enter:
            inst._enterstack.append(enter)  # noqa: SLF001
        return inst

    def __str__(self) -> str:
        """Generate querystring."""
        lastpart = '' if isinstance(self, Term) else ' #INCOMPLETE#'
        return ''.join([*self._parts, lastpart])

    def __getsql__(self) -> tuple[str, tuple[BindArg, ...]]:
        """Shorthand for str."""
        if self._calledon != Term:
            return self.endq().__getsql__()

        assert not self._enterstack, 'Enterstack should be empty'
        return str(self), tuple(self._args)

    @classmethod
    def _getfn(
        cls: type[QBase[T_co]],
        attr: Callable[Concatenate[QBase[T_co], P], R],
    ) -> Callable[Concatenate[QBase[T_co], P], R]:
        def func(obj: QBase[T_co], /, *args: P.args, **kwargs: P.kwargs) -> R:
            stack = obj._enterstack  # noqa: SLF001
            while stack:
                classes, callback = stack[-1]
                if cls not in classes:
                    break
                stack.pop()
                obj._parts += [callback]  # noqa: SLF001
            obj._calledon = cls  # noqa: SLF001
            return attr(obj, *args, **kwargs)

        return func

    @classmethod
    def __init_generics__(cls: type[QBase[T_co]]) -> None:
        """Set _next members."""
        module = sys.modules[cls.__module__].__dict__
        module.update({cls.__name__: cls, 'Type': type})
        try:
            for parent in get_generics(cls):
                origin = get_origin(parent)
                varmap = dict(
                    zip(
                        [
                            x
                            for base in getattr(origin, '__orig_bases__', [])
                            if get_origin(base) is Generic
                            for x in base.__parameters__
                        ],
                        get_args(parent),
                        strict=False,
                    ),
                )
                for key in [x for x in origin.__annotations__ if x.startswith('_next')]:
                    hints = get_type_hints(origin)
                    arg = get_args(hints[key])[0]
                    fref = varmap.get(arg, arg)

                    assert isinstance(fref, ForwardRef)
                    evald = fref._evaluate(  # noqa: SLF001
                        module,
                        module,
                        recursive_guard=frozenset(),
                    )
                    setattr(cls, key, evald)
        except NameError as err:
            name = str(err).split("'")[1]
            name = '' if name == cls.__name__ else name
            cls._unfinished.setdefault(
                (cls.__module__, name),
                [],
            ).append(cast('type[QBase[Texplicit]]', cls))
            return

        for unfikey in [(cls.__module__, cls.__name__), (cls.__module__, '')]:
            if unfikey in cls._unfinished:
                insts = cls._unfinished.pop(unfikey)
                for inst in insts:
                    inst.__init_generics__()

    def __init_subclass__(cls) -> None:
        """Add calledon wrapper."""
        super().__init_subclass__()
        for name in cls.__dict__:
            if name[0] == '_':
                continue
            attr = getattr(cls, name)
            assert callable(attr)
            setattr(cls, name, wraps(attr)(cls._getfn(attr)))
        cls.__init_generics__()


class Term(QBase[T_co]):
    """Noop."""

    def endq(self) -> Term[T_co]:
        """Done."""
        return self
